#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import functools
import json
import logging
from typing import Awaitable, Callable, Dict, Optional, Tuple, TypeVar, Union

import attr
from aiohttp import web
from multidict import CIMultiDictProxy

from matrix_content_scanner import logutils
from matrix_content_scanner.mcs_rust import crypto
from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.encrypted_file_metadata import (
    validate_encrypted_file_metadata,
)
from matrix_content_scanner.utils.errors import ContentScannerRestError
from matrix_content_scanner.utils.types import JsonDict

logger = logging.getLogger(__name__)

_next_request_seq = 0

_Handler = TypeVar("_Handler")


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _BytesResponse:
    """A binary response, and the headers to send back to the client alongside it."""

    headers: CIMultiDictProxy[str]
    content: bytes


def web_handler(
    func: Callable[
        [_Handler, web.Request], Awaitable[Tuple[int, Union[JsonDict, _BytesResponse]]]
    ],
) -> Callable[[_Handler, web.Request], Awaitable[web.Response]]:
    """Decorator that adds a wrapper to the given web handler method, which turns its
    return value into an aiohttp Response, and handles errors.

    Args:
        func: The function to wrap.

    Returns:
        The wrapper to run for this function.
    """

    def handle_error(status: int, reason: ErrCode, info: Optional[str]) -> web.Response:
        """Turns an error with the given parameters into an aiohttp Response.

        Args:
            status: The HTTP status code.
            reason: The error code to include in the response's JSON body.
            info: Optional extra info to include in the response's JSON body.
        """
        # Write the reason for the error into the response body, and add some extra info
        # if we have any.
        res_body: JsonDict = {"reason": reason}
        if info is not None:
            res_body["info"] = info

        res = _to_json_bytes(res_body)

        return web.Response(
            status=status,
            content_type="application/json",
            body=res,
        )

    @functools.wraps(func)
    async def wrapper(self: _Handler, request: web.Request) -> web.Response:
        """Run the wrapped method, and turn the return value into an aiohttp Response.

        If the wrapped method raises an exception, turn that into an aiohttp Response
        as well.

        Args:
            self: The object the wrapped method belongs to.
            request: The aiohttp Request to process.
        """
        # Set the request ID in the logging context, and increment the sequence for the
        # next request.
        global _next_request_seq
        request_id = f"{request.method}-{_next_request_seq}"
        logutils.set_request_id_in_context(request_id)
        _next_request_seq += 1

        # Check that the path is correct.
        if not request.path.startswith("/_matrix/media_proxy/unstable"):
            return handle_error(
                status=400,
                reason=ErrCode.UNKNOWN,
                info="Invalid path",
            )

        try:
            status, res = await func(self, request)

            # Set the response and headers according to the return value. If the handler
            # didn't return with a bytes response (in which it is responsible for
            # providing the headers, including the content-type one), default to json.
            headers: Union[Dict[str, str], CIMultiDictProxy[str]]
            if isinstance(res, _BytesResponse):
                raw_res = res.content
                headers = res.headers
            else:
                raw_res = _to_json_bytes(res)
                headers = {"content-type": "application/json"}

            return web.Response(
                status=status,
                body=raw_res,
                headers=headers,
            )
        except ContentScannerRestError as e:
            # If we get a REST error, use it to generate an error response.
            return handle_error(
                status=e.http_status,
                reason=e.reason,
                info=e.info,
            )
        except Exception as e:
            # Otherwise, just treat it as an unknown server error.
            logger.exception(e)
            return handle_error(
                status=500,
                reason=ErrCode.UNKNOWN,
                info="Internal Server Error",
            )

    return wrapper


def _to_json_bytes(content: JsonDict) -> bytes:
    """Converts a dict into JSON and encodes it to bytes."""
    return json.dumps(content).encode("UTF-8")


async def get_media_metadata_from_request(
    request: web.Request,
    crypto_handler: crypto.CryptoHandler,
) -> Tuple[str, JsonDict]:
    """Extracts, optionally decrypts, and validates encrypted file metadata from a
    request body.

    Args:
        request: The request to extract the data from.
        crypto_handler: The crypto handler to use if we need to decrypt an Olm-encrypted
            body.

    Raises:
        ContentScannerRestError(400) if the request's body is None or if the metadata
            didn't pass schema validation.
    """
    if request.content is None:
        raise ContentScannerRestError(
            400,
            ErrCode.MALFORMED_JSON,
            "No content in request body",
        )

    try:
        body = await request.json()
    except json.decoder.JSONDecodeError as e:
        raise ContentScannerRestError(400, ErrCode.MALFORMED_JSON, str(e))

    metadata = _metadata_from_body(body, crypto_handler)

    validate_encrypted_file_metadata(metadata)

    # Get the media path.
    url = metadata["file"]["url"]
    media_path = url[len("mxc://") :]

    return media_path, metadata


def _metadata_from_body(
    body: JsonDict, crypto_handler: crypto.CryptoHandler
) -> JsonDict:
    """Parse the given body as JSON, and decrypts it if needed.

    Args:
        body: The body, parsed as JSON.
        crypto_handler: The crypto handler to use if we need to decrypt an Olm-encrypted
            body.

    Returns:
        The parsed and decrypted file metadata.

    Raises:
        ContentScannerRestError(400) if the body isn't valid JSON or isn't a dictionary.
    """
    # Every POST request body in the API implemented by the content scanner is a dict.
    if not isinstance(body, dict):
        raise ContentScannerRestError(
            400,
            ErrCode.MALFORMED_JSON,
            "Body must be a dictionary",
        )

    # Check if the metadata is encrypted, if not then the metadata is in clear text in
    # the body so just return it.
    encrypted_body: Optional[JsonDict] = body.get("encrypted_body")
    if encrypted_body is None:
        return body

    # If it is encrypted, decrypt it and return the decrypted version.
    try:
        decrypted: JsonDict = json.loads(
            crypto_handler.decrypt_body(
                ciphertext=encrypted_body["ciphertext"],
                mac=encrypted_body["mac"],
                ephemeral=encrypted_body["ephemeral"],
            )
        )
        return decrypted
    except Exception as e:
        logger.exception("Failed to decrypt encrypted body")
        raise ContentScannerRestError(
            http_status=400,
            reason=ErrCode.FAILED_TO_DECRYPT,
            info=str(e),
        )



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import Any, Dict, List, Optional, Union

import attr
import humanfriendly
from jsonschema import ValidationError, validate

from matrix_content_scanner.utils.errors import ConfigError

_ONE_WEEK_SECONDS = 604800.0


def _parse_duration(duration: Optional[Union[str, float]]) -> Optional[float]:
    """Parse a time duration into a float representing an amount of second. If the given
    value is None, or already a float, returns it as is.

    Args:
        duration: The duration to parse.

    Returns:
        The number of seconds in the given duration.
    """
    if duration is None or isinstance(duration, float):
        return duration

    try:
        return humanfriendly.parse_timespan(duration)
    except humanfriendly.InvalidTimespan as e:
        raise ConfigError(e)


def _parse_size(size: Optional[Union[str, float]]) -> Optional[float]:
    """Parse a file size into a float representing the number of bytes for that size. If
    the given value is None, or already a float, returns it as is.

    Args:
        size: The size to parse.

    Returns:
        The number of bytes represented by the given size.
    """
    if size is None or isinstance(size, float):
        return size

    try:
        return humanfriendly.parse_size(size)
    except humanfriendly.InvalidSize as e:
        raise ConfigError(e)


# Schema to validate the raw configuration dictionary against.
_config_schema = {
    "type": "object",
    "required": ["web", "scan", "crypto"],
    "additionalProperties": False,
    "properties": {
        "web": {
            "type": "object",
            "required": ["host", "port"],
            "additionalProperties": False,
            "properties": {
                "host": {"type": "string"},
                "port": {"type": "integer"},
            },
        },
        "scan": {
            "type": "object",
            "required": ["script", "temp_directory"],
            "additionalProperties": False,
            "properties": {
                "script": {"type": "string"},
                "temp_directory": {"type": "string"},
                "removal_command": {"type": "string"},
                "allowed_mimetypes": {"type": "array", "items": {"type": "string"}},
                "blocked_mimetypes": {"type": "array", "items": {"type": "string"}},
            },
        },
        "download": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "base_homeserver_url": {"type": "string"},
                "proxy": {"type": "string"},
                "additional_headers": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
        },
        "crypto": {
            "type": "object",
            "required": ["pickle_path", "pickle_key"],
            "additionalProperties": False,
            "properties": {
                "pickle_path": {"type": "string"},
                "pickle_key": {"type": "string"},
            },
        },
        "result_cache": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "max_size": {"type": "integer"},
                "ttl": {"type": ["string", "number"]},
                "exit_codes_to_ignore": {
                    "type": "array",
                    "items": {"type": "integer"},
                },
                "max_file_size": {"type": ["string", "number"]},
            },
        },
    },
}


@attr.s(auto_attribs=True, frozen=True)
class WebConfig:
    """Configuration for serving the HTTP API."""

    host: str
    port: int


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ScanConfig:
    """Configuration for scanning files."""

    script: str
    temp_directory: str
    removal_command: str = "rm"
    allowed_mimetypes: Optional[List[str]] = None
    blocked_mimetypes: Optional[List[str]] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ResultCacheConfig:
    """Configuration for caching scan results."""

    max_size: int = 1024
    ttl: float = attr.ib(default=_ONE_WEEK_SECONDS, converter=_parse_duration)
    exit_codes_to_ignore: Optional[List[int]] = None
    max_file_size: Optional[float] = attr.ib(default=None, converter=_parse_size)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class DownloadConfig:
    """Configuration for downloading files."""

    base_homeserver_url: Optional[str] = None
    proxy: Optional[str] = None
    additional_headers: Optional[Dict[str, str]] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CryptoConfig:
    """Configuration for decrypting encrypted bodies."""

    pickle_path: str
    pickle_key: str


class MatrixContentScannerConfig:
    def __init__(self, config_dict: Dict[str, Any]):
        if not isinstance(config_dict, dict):
            raise ConfigError("Bad configuration format")

        try:
            validate(config_dict, _config_schema)
        except ValidationError as e:
            raise ConfigError(e.message)

        self.web = WebConfig(**(config_dict.get("web") or {}))
        self.scan = ScanConfig(**(config_dict.get("scan") or {}))
        self.crypto = CryptoConfig(**(config_dict.get("crypto") or {}))
        self.download = DownloadConfig(**(config_dict.get("download") or {}))
        self.result_cache = ResultCacheConfig(**(config_dict.get("result_cache") or {}))

        # Don't allow both allowlist and blocklist for MIME types, since we do not document
        # the semantics for that and it is in any case pointless.
        # This could have been expressed in JSONSchema but I suspect the error message would be poor
        # in that case.
        if (
            self.scan.allowed_mimetypes is not None
            and self.scan.blocked_mimetypes is not None
        ):
            raise ConfigError(
                "Both `scan.allowed_mimetypes` and `scan.blocked_mimetypes` are specified, which is not allowed!"
            )



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import json
from typing import Dict, List, Optional, Tuple, Union
from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock, call

from multidict import CIMultiDict, CIMultiDictProxy, MultiDictProxy

from matrix_content_scanner.utils.errors import (
    ContentScannerRestError,
    WellKnownDiscoveryError,
)
from matrix_content_scanner.utils.types import JsonDict

from tests.testutils import (
    MEDIA_PATH,
    SMALL_PNG,
    get_base_media_headers,
    get_content_scanner,
    to_thumbnail_params,
)


class FileDownloaderTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        # Set a fixed base URL so that .well-known discovery doesn't get in the way.
        content_scanner = get_content_scanner(
            {"download": {"base_homeserver_url": "http://my-site.com"}}
        )
        self.downloader = content_scanner.file_downloader

        self.media_status = 200
        self.media_body = SMALL_PNG
        self.media_headers = get_base_media_headers()

        async def _get(
            url: str,
            query: Optional[MultiDictProxy[str]] = None,
            auth_header: Optional[str] = None,
        ) -> Tuple[int, bytes, CIMultiDictProxy[str]]:
            """Mock for the _get method on the file downloader that doesn't serve a
            .well-known client file.
            """
            if (
                url.endswith(
                    (
                        "/_matrix/media/v3/download/" + MEDIA_PATH,
                        "/_matrix/media/r0/download/" + MEDIA_PATH,
                    )
                )
                or "/_matrix/media/v3/thumbnail/" + MEDIA_PATH in url
                or "/_matrix/media/r0/thumbnail/" + MEDIA_PATH in url
            ):
                return self.media_status, self.media_body, self.media_headers
            if (
                url.endswith(("/_matrix/client/v1/media/download/" + MEDIA_PATH,))
                or "/_matrix/client/v1/media/thumbnail/" + MEDIA_PATH in url
            ):
                if auth_header is not None:
                    return self.media_status, self.media_body, self.media_headers
                else:
                    return 404, b"Not found", CIMultiDictProxy(CIMultiDict())
            elif url.endswith("/.well-known/matrix/client"):
                return 404, b"Not found", CIMultiDictProxy(CIMultiDict())

            raise RuntimeError("Unexpected request on %s" % url)

        # Mock _get so we don't actually try to download files.
        self.get_mock = Mock(side_effect=_get)
        self.downloader._get = self.get_mock  # type: ignore[method-assign]

    async def test_download(self) -> None:
        """Tests that downloading a file works."""
        media = await self.downloader.download_file(MEDIA_PATH)
        self.assertEqual(media.content, SMALL_PNG)
        self.assertEqual(media.content_type, "image/png")

        # Check that we tried downloading from the set base URL.
        args = self.get_mock.call_args.args
        self.assertTrue(args[0].startswith("http://my-site.com/"))

    async def test_download_auth_media(self) -> None:
        """Tests that downloading a file works using authenticated media."""
        media = await self.downloader.download_file(
            MEDIA_PATH, auth_header="Bearer access_token"
        )
        self.assertEqual(media.content, SMALL_PNG)
        self.assertEqual(media.content_type, "image/png")

        # Check that we tried downloading from the set base URL.
        args = self.get_mock.call_args.args
        self.assertTrue(args[0].startswith("http://my-site.com/"))
        self.assertIn("/_matrix/client/v1/media/download/" + MEDIA_PATH, args[0])

    async def test_download_auth_media_invalid_token(self) -> None:
        """Tests that downloading an authenticated media file with an invalid access
        token returns the correct error code.
        """
        self.media_status = 401
        self.media_body = (
            b'{"errcode":"M_UNKNOWN_TOKEN","error":"Invalid access token"}'
        )
        self._set_headers({"content-type": ["application/json"]})

        # Check that we fail at downloading the file.
        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(
                MEDIA_PATH, auth_header="Bearer access_token"
            )

        self.assertEqual(cm.exception.http_status, 401)
        self.assertEqual(cm.exception.reason, "M_UNKNOWN_TOKEN")

        # Check that we tried downloading from the set base URL.
        args = self.get_mock.call_args.args
        self.assertTrue(args[0].startswith("http://my-site.com/"))
        self.assertIn("/_matrix/client/v1/media/download/" + MEDIA_PATH, args[0])

    async def test_download_auth_media_missing_token(self) -> None:
        """Tests that downloading an authenticated media file with a missing access
        token returns the correct error code.
        """
        self.media_status = 401
        self.media_body = (
            b'{"errcode":"M_MISSING_TOKEN","error":"Missing access token"}'
        )
        self._set_headers({"content-type": ["application/json"]})

        # Check that we fail at downloading the file.
        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(
                MEDIA_PATH, auth_header="Bearer access_token"
            )

        self.assertEqual(cm.exception.http_status, 401)
        self.assertEqual(cm.exception.reason, "M_MISSING_TOKEN")

        # Check that we tried downloading from the set base URL.
        args = self.get_mock.call_args.args
        self.assertTrue(args[0].startswith("http://my-site.com/"))
        self.assertIn("/_matrix/client/v1/media/download/" + MEDIA_PATH, args[0])

    async def test_no_base_url(self) -> None:
        """Tests that configuring a base homeserver URL means files are downloaded from
        that homeserver (rather than the one the files were uploaded to) and .well-known
        discovery is bypassed.
        """
        self.downloader._base_url = None
        await self.downloader.download_file(MEDIA_PATH)

        # Check that we've tried making a .well-known discovery request before
        # downloading the file.
        self.assertEqual(self.get_mock.call_count, 2)
        self.assertEqual(
            self.get_mock.mock_calls[0], call("https://foo/.well-known/matrix/client")
        )
        self.assertEqual(
            self.get_mock.mock_calls[1],
            call(
                "https://foo/_matrix/media/v3/download/" + MEDIA_PATH,
                query=None,
                auth_header=None,
            ),
        )

    async def test_retry_on_404(self) -> None:
        """Tests that if we get a 404 when trying to download a file on a v3 path, we
        retry with an r0 path for backwards compatibility.
        """
        self.media_status = 404
        self.media_body = b"Not found"
        self._set_headers({"content-type": ["text/plain"]})

        await self._test_retry()

    async def test_retry_on_unrecognised(self) -> None:
        """Tests that if we get a Synapse-style M_UNRECOGNIZED response when trying to
        download a file on a v3 path, we retry with an r0 path for backwards
        compatibility.
        """
        self.media_status = 400
        self.media_body = b'{"errcode":"M_UNRECOGNIZED","error":"Unrecognized request"}'
        self._set_headers({"content-type": ["application/json"]})

        await self._test_retry()

    async def _test_retry(self) -> None:
        """Tests that in a set specific case a failure to download a file from a v3
        download path means we retry the request on an r0 one for backwards compatibility.
        """
        # Check that we eventually fail at downloading the file.
        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(MEDIA_PATH)

        self.assertEqual(cm.exception.http_status, 404)
        self.assertEqual(cm.exception.info, "File not found")

        # Check that we sent out two requests: one to the v3 path and one to the r0 path.
        self.assertEqual(self.get_mock.call_count, 2)
        self.assertEqual(
            self.get_mock.mock_calls[0],
            call(
                "http://my-site.com/_matrix/media/v3/download/" + MEDIA_PATH,
                query=None,
                auth_header=None,
            ),
        )
        self.assertEqual(
            self.get_mock.mock_calls[1],
            call(
                "http://my-site.com/_matrix/media/r0/download/" + MEDIA_PATH,
                query=None,
                auth_header=None,
            ),
        )

    async def test_no_retry(self) -> None:
        """Tests that in a set specific case a failure to download a file from a v1
        authenticated media download path means we don't retry the request.
        """
        self.media_status = 400
        self.media_body = b'{"errcode":"M_UNRECOGNIZED","error":"Unrecognized request"}'
        self._set_headers({"content-type": ["application/json"]})

        # Check that we eventually fail at downloading the file.
        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(
                MEDIA_PATH, auth_header="Bearer access_token"
            )

        self.assertEqual(cm.exception.http_status, 404)
        self.assertEqual(cm.exception.info, "File not found")

        # Check that we sent out only one request.
        self.assertEqual(self.get_mock.call_count, 1)
        self.assertEqual(
            self.get_mock.mock_calls[0],
            call(
                "http://my-site.com/_matrix/client/v1/media/download/" + MEDIA_PATH,
                query=None,
                auth_header="Bearer access_token",
            ),
        )

    async def test_thumbnail(self) -> None:
        """Tests that we can download a thumbnail and that the parameters to generate the
        thumbnail are correctly passed on to the homeserver.
        """
        await self.downloader.download_file(
            MEDIA_PATH, to_thumbnail_params({"height": "50"})
        )

        url: str = self.get_mock.call_args.args[0]
        query: CIMultiDictProxy[str] = self.get_mock.call_args.kwargs["query"]
        self.assertIn("/thumbnail/", url)
        self.assertIn("height", query)
        self.assertEqual(query.get("height"), "50", query.getall("height"))

    async def test_thumbnail_auth_media(self) -> None:
        """Tests that we can download a thumbnail and that the parameters to generate the
        thumbnail are correctly passed on to the homeserver using authenticated media.
        """
        await self.downloader.download_file(
            MEDIA_PATH, to_thumbnail_params({"height": "50"}), "Bearer access_token"
        )

        url: str = self.get_mock.call_args.args[0]
        query: CIMultiDictProxy[str] = self.get_mock.call_args.kwargs["query"]
        self.assertIn("/thumbnail/", url)
        self.assertIn("/_matrix/client/v1/media/thumbnail/" + MEDIA_PATH, url)
        self.assertIn("height", query)
        self.assertEqual(query.get("height"), "50", query.getall("height"))

    async def test_multiple_content_type(self) -> None:
        """Tests that we raise an error if the homeserver responds with too many
        Content-Type headers.
        """
        self._set_headers({"content-type": ["image/jpeg", "image/png"]})

        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(MEDIA_PATH)

        self.assertEqual(cm.exception.http_status, 502)
        assert cm.exception.info is not None
        self.assertTrue("Content-Type" in cm.exception.info)

    async def test_no_content_type(self) -> None:
        """Tests that we raise an error if the homeserver responds with no Content-Type
        headers.
        """
        self._set_headers({})

        with self.assertRaises(ContentScannerRestError) as cm:
            await self.downloader.download_file(MEDIA_PATH)

        self.assertEqual(cm.exception.http_status, 502)
        assert cm.exception.info is not None
        self.assertTrue("Content-Type" in cm.exception.info)

    def _set_headers(self, headers: Dict[str, List[str]]) -> None:
        """Replace the headers set in setUp with ones constructed from the provided
        dictionary.

        Args:
            headers: The raw headers to set.
        """
        md: CIMultiDict[str] = CIMultiDict()
        for k, v in headers.items():
            for el in v:
                md.add(k, el)

        self.media_headers = CIMultiDictProxy(md)


class WellKnownDiscoveryTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.downloader = get_content_scanner().file_downloader

        self.well_known_status = 200
        self.well_known_body: Union[bytes, JsonDict] = b""

        self.versions_status = 200

        async def _get(
            url: str,
            query: Optional[MultiDictProxy[str]] = None,
            auth_header: Optional[str] = None,
        ) -> Tuple[int, bytes, CIMultiDictProxy[str]]:
            """Mock for the _get method on the file downloader that serves a .well-known
            client file.
            """
            if url.endswith("/.well-known/matrix/client"):
                if isinstance(self.well_known_body, bytes):
                    body_bytes = self.well_known_body
                else:
                    body_bytes = json.dumps(self.well_known_body).encode("utf-8")

                return (
                    self.well_known_status,
                    body_bytes,
                    CIMultiDictProxy(CIMultiDict()),
                )
            elif url.endswith("/_matrix/client/versions"):
                return self.versions_status, b"{}", CIMultiDictProxy(CIMultiDict())
            elif url.endswith("/_matrix/media/v3/download/" + MEDIA_PATH):
                return 200, SMALL_PNG, get_base_media_headers()

            raise RuntimeError("Unexpected request on %s" % url)

        # Mock _get so we don't actually try to download files.
        self.get_mock = Mock(side_effect=_get)
        self.downloader._get = self.get_mock  # type: ignore[method-assign]

    async def test_discover(self) -> None:
        """Checks that the base URL to use to download files can be discovered via
        .well-known discovery.
        """
        self.well_known_body = {"m.homeserver": {"base_url": "https://foo.bar"}}

        await self.downloader.download_file(MEDIA_PATH)

        # Check that we got 3 calls:
        #  * one to retrieve the .well-known file
        #  * one to check that the base URL can be used to interact with a homeserver
        #    (by hitting the /_matrix/client/versions endpoint)
        #  * one to download the file
        self.assertEqual(self.get_mock.call_count, 3, self.get_mock.mock_calls)

        calls = self.get_mock.mock_calls

        self.assertEqual(calls[0], call("https://foo/.well-known/matrix/client"))
        self.assertTrue(calls[1], call("https://foo.bar/_matrix/client/versions"))
        self.assertTrue(
            calls[2], call("https://foo.bar/_matrix/media/v3/download/" + MEDIA_PATH)
        )

    async def test_error_status(self) -> None:
        """Tests that we raise a WellKnownDiscoveryError if the server responded with an
        error."""
        self.well_known_status = 401
        await self._assert_discovery_fail()

    async def test_malformed_content(self) -> None:
        """Tests that we raise a WellKnownDiscoveryError if the server responded with a
        body that isn't compliant with the Matrix specification."""
        self.well_known_body = {"m.homeserver": "https://foo.bar"}
        await self._assert_discovery_fail()

    async def test_not_valid_homeserver(self) -> None:
        """Tests that we raise a WellKnownDiscoveryError if the server at the provided
        base URL isn't a Matrix homeserver."""
        self.versions_status = 404
        await self._assert_discovery_fail()

    async def test_404_no_fail(self) -> None:
        """Tests that we don't raise a WellKnownDiscoveryError if the .well-known file
        couldn't be found, and that we return None instead of the discovered base URL in
        this case.
        """
        self.well_known_status = 404
        res = await self.downloader._discover_via_well_known("foo")
        self.assertIsNone(res)

    async def _assert_discovery_fail(self) -> None:
        """Checks that .well-known discovery fails and raises a WellKnownDiscoveryError."""
        with self.assertRaises(WellKnownDiscoveryError):
            await self.downloader._discover_via_well_known("foo")



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import asyncio
import hashlib
import logging
import os
import subprocess
from asyncio import Future
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import attr
import magic
from cachetools import TTLCache
from canonicaljson import encode_canonical_json
from humanfriendly import format_size
from multidict import MultiMapping

from matrix_content_scanner.mcs_rust import crypto
from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import (
    ContentScannerRestError,
    FileDirtyError,
    FileMimeTypeForbiddenError,
)
from matrix_content_scanner.utils.types import JsonDict, MediaDescription

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, frozen=True)
class CacheEntry:
    """An entry in the scanner's result cache."""

    # The result of the scan: True if the scan passed, False otherwise.
    result: bool

    # The media that was scanned, so we can return it in future requests. We only cache
    # it if the scan succeeded and the file's size does not exceed the configured limit,
    # otherwise it's None.
    media: Optional[MediaDescription] = None

    # Hash of the media content, so we can make sure no malicious servers changed the file
    # since we've scanned it (e.g. if we need to re-download it because the file was too
    # big). None if the scan failed.
    media_hash: Optional[str] = None

    # Info to include in the FileDirtyError if the scan failed.
    info: Optional[str] = None


class Scanner:
    def __init__(self, mcs: "MatrixContentScanner"):
        self._file_downloader = mcs.file_downloader
        self._script = mcs.config.scan.script
        self._removal_command = mcs.config.scan.removal_command
        self._store_directory = Path(mcs.config.scan.temp_directory).resolve(
            strict=True
        )

        # Result cache settings.
        self._result_cache: TTLCache[str, CacheEntry] = TTLCache(
            maxsize=mcs.config.result_cache.max_size,
            ttl=mcs.config.result_cache.ttl,
        )

        if mcs.config.result_cache.exit_codes_to_ignore is None:
            self._exit_codes_to_ignore = []
        else:
            self._exit_codes_to_ignore = mcs.config.result_cache.exit_codes_to_ignore

        self._max_size_to_cache = mcs.config.result_cache.max_file_size

        # List of MIME types we should allow.
        # If None, we fall back to `_blocked_mimetypes`.
        # If that's also None, we don't fail files based on their
        # MIME types (besides comparing it with the Content-Type header from the server
        # for unencrypted files).
        self._allowed_mimetypes = mcs.config.scan.allowed_mimetypes

        # List of MIME types we should block.
        # Must not be specified at the same time as `_allowed_mimetypes`.
        # See the comment for `_allowed_mimetypes` for the semantics.
        self._blocked_mimetypes = mcs.config.scan.blocked_mimetypes

        # Cache of futures for files that are currently scanning and downloading, so that
        # concurrent requests don't cause a file to be downloaded and scanned twice.
        self._current_scans: Dict[str, Future[MediaDescription]] = {}

        # Limit the number of concurrent scans.
        self._current_scan_semaphore = asyncio.Semaphore(100)

    async def scan_file(
        self,
        media_path: str,
        metadata: Optional[JsonDict] = None,
        thumbnail_params: Optional["MultiMapping[str]"] = None,
        auth_header: Optional[str] = None,
    ) -> MediaDescription:
        """Download and scan the given media.

        Unless the scan fails with one of the codes listed in `do_not_cache_exit_codes`,
        also cache the result.

        If the file already has an entry in the result cache, return this value without
        downloading the file again (unless we purposefully did not cache the file's
        content to save up on memory).

        If a file is currently already being downloaded or scanned as a result of another
        request, don't download it again and use the result from the first request.

        Args:
            media_path: The `server_name/media_id` path for the media.
            metadata: The metadata attached to the file (e.g. decryption key), or None if
                the file isn't encrypted.
            thumbnail_params: If present, then we want to request and scan a thumbnail
                generated with the provided parameters instead of the full media.
            auth_header: If present, we forward the given Authorization header, this is
                required for authenticated media endpoints.

        Returns:
            A description of the media.

        Raises:
            ContentScannerRestError if the file could not be downloaded.
            FileDirtyError if the result of the scan said that the file is dirty, or if
                the media path is malformed.
        """
        # Compute the key to use when caching, both in the current scans cache and in the
        # results cache.
        cache_key = self._get_cache_key_for_file(media_path, metadata, thumbnail_params)
        if cache_key not in self._current_scans:
            # Create a future in the context of the current event loop.
            loop = asyncio.get_event_loop()
            f = loop.create_future()
            # Register the future in the current scans cache so that subsequent queries
            # can use it.
            self._current_scans[cache_key] = f
            # Try to download and scan the file.
            try:
                res = await self._scan_file(
                    cache_key, media_path, metadata, thumbnail_params, auth_header
                )
                # Set the future's result, and mark it as done.
                f.set_result(res)
                # Return the result.
                return res
            except Exception as e:
                # If there's an exception, catch it, pass it on to the future, and raise
                # it.
                f.set_exception(e)
                # We retrieve the exception from the future, because if we don't and no
                # other request is awaiting on the future, asyncio complains about "Future
                # exception was never retrieved".
                f.exception()
                raise
            finally:
                # Remove the future from the cache.
                del self._current_scans[cache_key]

        return await self._current_scans[cache_key]

    async def _scan_file(
        self,
        cache_key: str,
        media_path: str,
        metadata: Optional[JsonDict] = None,
        thumbnail_params: Optional[MultiMapping[str]] = None,
        auth_header: Optional[str] = None,
    ) -> MediaDescription:
        """Download and scan the given media.

        Unless the scan fails with one of the codes listed in `do_not_cache_exit_codes`,
        also cache the result.

        If the file already has an entry in the result cache, return this value without
        downloading the file again (unless we purposefully did not cache the file's
        content to save up on memory).

        Args:
            cache_key: The key to use to cache the result of the scan in the result cache.
            media_path: The `server_name/media_id` path for the media.
            metadata: The metadata attached to the file (e.g. decryption key), or None if
                the file isn't encrypted.
            thumbnail_params: If present, then we want to request and scan a thumbnail
                generated with the provided parameters instead of the full media.
            auth_header: If present, we forward the given Authorization header, this is
                required for authenticated media endpoints.

        Returns:
            A description of the media.

        Raises:
            ContentScannerRestError if the file could not be downloaded.
            FileDirtyError if the result of the scan said that the file is dirty, or if
                the media path is malformed.
        """
        # The media to scan.
        media: Optional[MediaDescription] = None

        # Return the cached result if there's one.
        cache_entry = self._result_cache.get(cache_key)
        if cache_entry is not None:
            logger.info("Found a cached result %s", cache_entry.result)

            if cache_entry.result is False:
                # Feed the additional info we might have added when caching the error,
                # into the new error.
                raise FileDirtyError(info=cache_entry.info)

            if cache_entry.media is not None:
                return cache_entry.media

            # If we don't have the media cached
            logger.info(
                "Got a positive result from cache without a media, downloading file",
            )

            media = await self._file_downloader.download_file(
                media_path=media_path,
                thumbnail_params=thumbnail_params,
                auth_header=auth_header,
            )

            # Compare the media's hash to ensure the server hasn't changed the file since
            # the last scan. If it has changed, shout about it in the logs, discard the
            # cache entry and scan it again.
            media_hash = hashlib.sha256(media.content).hexdigest()
            if media_hash == cache_entry.media_hash:
                return media

            logger.warning(
                "Media has changed since last scan (cached hash: %s, new hash: %s),"
                " discarding cached result and scanning again",
                cache_entry.media_hash,
                media_hash,
            )

            del self._result_cache[cache_key]

        # Check if the media path is valid and only contains one slash (otherwise we'll
        # have issues parsing it further down the line).
        if media_path.count("/") != 1:
            info = "Malformed media ID"
            self._result_cache[cache_key] = CacheEntry(
                result=False,
                info=info,
            )
            raise FileDirtyError(info)

        # Download the file if we don't already have it.
        if media is None:
            media = await self._file_downloader.download_file(
                media_path=media_path,
                thumbnail_params=thumbnail_params,
                auth_header=auth_header,
            )

        # Download and scan the file.
        try:
            media, cacheable = await self._scan_media(media, media_path, metadata)
        except FileDirtyError as e:
            if e.cacheable:
                logger.info("Caching scan failure")

                # If the test fails, don't store the media to save memory.
                self._result_cache[cache_key] = CacheEntry(
                    result=False,
                    media=None,
                    info=e.info,
                )

            raise

        # Update the cache if the result should be cached.
        if cacheable:
            logger.info("Caching scan success")

            cached_media: Optional[MediaDescription] = media

            if (
                self._max_size_to_cache is not None
                and len(media.content) > self._max_size_to_cache
            ):
                # Don't cache the file's content if it exceeds the maximum allowed file
                # size, to minimise memory usage.
                logger.info(
                    "File content has size %s, which is more than %s, not caching content",
                    format_size(len(media.content)),
                    format_size(self._max_size_to_cache),
                )

                cached_media = None

            # Hash the media, that way if we need to re-download the file we can make sure
            # it's the right one. We get a hex digest in case we want to print it later.
            media_hash = hashlib.sha256(media.content).hexdigest()

            self._result_cache[cache_key] = CacheEntry(
                result=True,
                media=cached_media,
                media_hash=media_hash,
            )

        return media

    async def _scan_media(
        self,
        media: MediaDescription,
        media_path: str,
        metadata: Optional[JsonDict] = None,
    ) -> Tuple[MediaDescription, bool]:
        """Scans the given media.

        Args:
            media: The already downloaded media. If provided, the download step is
                skipped. Usually provided if we've re-downloaded a file with a cached
                result, but the file changed since the initial scan.
            media_path: The `server_name/media_id` path for the media.
            metadata: The metadata attached to the file (e.g. decryption key), or None if
                the file isn't encrypted.

        Returns:
            A description of the media, as well as a boolean indicating whether the
            successful scan result should be cached or not.

        Raises:
            FileDirtyError if the result of the scan said that the file is dirty, or if
                the media path is malformed.
        """

        # Decrypt the content if necessary.
        media_content = media.content
        if metadata is not None:
            # If the file is encrypted, we need to decrypt it before we can scan it.
            media_content = self._decrypt_file(media_content, metadata)

        # Check the file's MIME type to see if it's allowed.
        self._check_mimetype(media_content)

        # Write the file to disk.
        file_path = self._write_file_to_disk(media_path, media_content)

        # Scan the file and see if the result is positive or negative.
        exit_code = await self._run_scan(file_path)
        result = exit_code == 0

        # If the exit code isn't part of the ones we should ignore, cache the result.
        cacheable = True
        if exit_code in self._exit_codes_to_ignore:
            logger.info(
                "Scan returned exit code %d which must not be cached", exit_code
            )
            cacheable = False

        # Delete the file now that we've scanned it.
        logger.info("Scan has finished, removing file")
        removal_command_parts = self._removal_command.split()
        removal_command_parts.append(file_path)
        subprocess.run(removal_command_parts)

        # Raise an error if the result isn't clean.
        if result is False:
            raise FileDirtyError(cacheable=cacheable)

        return media, cacheable

    def _get_cache_key_for_file(
        self,
        media_path: str,
        metadata: Optional[JsonDict],
        thumbnail_params: Optional[MultiMapping[str]],
    ) -> str:
        """Generates the key to use to store the result for the given media in the result
        cache.

        The key is computed using the media's `server_name/media_id` path, but also the
        metadata dict (stringified), in case e.g. the decryption key changes, as well as
        the parameters used to generate the thumbnail if any (stringified), to
        differentiate thumbnails from full-sized media.
        The resulting key is a sha256 hash of the concatenation of these two values.

        Args:
            media_path: The `server_name/media_id` path of the file to scan.
            metadata: The file's metadata (or None if the file isn't encrypted).
            thumbnail_params: The parameters to generate thumbnail with. If no parameter
                is passed, this will be an empty dict. If the media being requested is not
                a thumbnail, this will be None.
        """
        # If we're provided with thumbnailing parameters, turn them into a structure that
        # can be serialised as JSON.
        thumbnail_params_json: Optional[Dict[str, List[str]]] = None
        if thumbnail_params is not None:
            thumbnail_params_json = {}
            for k in thumbnail_params.keys():
                thumbnail_params_json[k] = thumbnail_params.getall(k)

        hash = hashlib.sha256()
        hash.update(media_path.encode("utf8"))
        hash.update(b"\0")
        hash.update(encode_canonical_json(metadata))
        hash.update(b"\0")
        hash.update(encode_canonical_json(thumbnail_params_json))

        return hash.hexdigest()

    def _decrypt_file(self, body: bytes, metadata: JsonDict) -> bytes:
        """Extract decryption information from the file's metadata and decrypt it.

        Args:
            body: The encrypted body of the file.
            metadata: The part of the request that includes decryption information.

        Returns:
            The decrypted content of the file.

        Raises:
            ContentScannerRestError(400) if the decryption failed.
        """
        logger.info("Decrypting encrypted file")

        # Decrypt the file.
        try:
            return crypto.decrypt_attachment(body, metadata["file"])
        except Exception as e:
            raise ContentScannerRestError(
                http_status=400,
                reason=ErrCode.FAILED_TO_DECRYPT,
                info=str(e),
            )

    def _write_file_to_disk(self, media_path: str, body: bytes) -> str:
        """Writes the given content to disk. The final file name will be a concatenation
        of `temp_directory` and the media's `server_name/media_id` path.

        Args:
            media_path: The `server_name/media_id` path of the media we're processing.
            body: The bytes to write to disk.

        Returns:
            The full path to the newly written file.

        Raises:
            FileDirtyError if the media path is malformed in a way that would cause the
                file to be written outside the configured directory.
        """
        # Figure out the full absolute path for this file.
        full_path = self._store_directory.joinpath(media_path).resolve()
        try:
            # Check if the full path is a sub-path to the store's path, to make sure
            # there isn't any '..' etc. in the full path, which would cause us to try
            # writing outside the store's directory.
            full_path.relative_to(self._store_directory)
        except ValueError:
            raise FileDirtyError("Malformed media ID")

        logger.info("Writing file to %s", full_path)

        # Create any directory we need.
        os.makedirs(full_path.parent, exist_ok=True)

        with open(full_path, "wb") as fp:
            fp.write(body)

        return str(full_path)

    async def _run_scan(self, file_name: str) -> int:
        """Runs the scan script, passing it the given file name.

        Args:
            file_name: Name of the file to scan.

        Returns:
            The exit code the script returned.
        """
        async with self._current_scan_semaphore:
            process = await asyncio.create_subprocess_exec(
                self._script, file_name, stderr=asyncio.subprocess.PIPE
            )
            _, stderr = await process.communicate()
            retcode = await process.wait()
            if retcode == 0:
                logger.info("Scan succeeded")
            else:
                logger.info(
                    "Scanning failed with exit code %d. Stderr: %s",
                    retcode,
                    stderr.decode(),
                )

            return retcode

    def _check_mimetype(self, media_content: bytes) -> None:
        """Detects the MIME type of the provided bytes, and checks that this type is allowed
        (if an allow list is provided in the configuration)
        Args:
            media_content: The file's content. If the file is encrypted, this is its
                decrypted content.
        Raises:
            FileDirtyError if one of the checks fail.
        """
        detected_mimetype = magic.from_buffer(media_content, mime=True)
        logger.debug("Detected MIME type for file is %s", detected_mimetype)

        # If there's an allow list for MIME types, check that the MIME type that's been
        # detected for this file is in it.
        if (
            self._allowed_mimetypes is not None
            and detected_mimetype not in self._allowed_mimetypes
        ):
            logger.error(
                "MIME type for file is forbidden: %s",
                detected_mimetype,
            )
            raise FileMimeTypeForbiddenError(
                f"File type: {detected_mimetype} not allowed"
            )

        # If there's a block list for MIME types, check that the MIME type detected for
        # this file is NOT in it.
        if (
            self._blocked_mimetypes is not None
            and detected_mimetype in self._blocked_mimetypes
        ):
            logger.error(
                "MIME type for file is forbidden: %s",
                detected_mimetype,
            )
            raise FileMimeTypeForbiddenError(
                f"File type: {detected_mimetype} not allowed"
            )



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import TYPE_CHECKING, Tuple

from aiohttp import web

from matrix_content_scanner.servlets import web_handler
from matrix_content_scanner.utils.types import JsonDict

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner


class PublicKeyHandler:
    def __init__(self, content_scanner: "MatrixContentScanner") -> None:
        self._crypto_handler = content_scanner.crypto_handler

    @web_handler
    async def handle_public_key(self, request: web.Request) -> Tuple[int, JsonDict]:
        """Handles GET requests to .../public_key"""
        return 200, {"public_key": self._crypto_handler.public_key}



from matrix_content_scanner.utils.types import JsonDict

class CryptoHandler:
    def __init__(self, pickle_key: str, pickle_path: str) -> None: ...
    @property
    def public_key(self) -> str: ...
    def decrypt_body(self, ciphertext: str, mac: str, ephemeral: str) -> str: ...
    def encrypt(self, public_key: str, payload: str) -> PkMessage: ...

class PkMessage:
    @property
    def ephemeral_key(self) -> str: ...
    @property
    def mac(self) -> str: ...
    @property
    def ciphertext(self) -> str: ...

def decrypt_attachment(body: bytes, key_info: JsonDict) -> bytes: ...



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import logging
from contextvars import ContextVar
from typing import Any

# The request's ID.
request_id: ContextVar[str] = ContextVar("request_id")


def setup_custom_factory() -> None:
    """Generates a new record factory, chained to the current factory, and sets it as the
    new default record factory.

    The new factory adds attributes for the media path and request type to log records,
    and populates them using the matching ContextVars;
    """
    old_factory = logging.getLogRecordFactory()

    def _factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
        record = old_factory(*args, **kwargs)
        record.request_id = request_id.get(None)
        return record

    logging.setLogRecordFactory(_factory)


def set_request_id_in_context(v: str) -> None:
    """Sets the request_id ContextVar to the given value.

    Args:
        v: The value to set the ContextVar.
    """
    request_id.set(v)



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import json
import unittest

from matrix_content_scanner.servlets import _metadata_from_body
from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import ContentScannerRestError
from matrix_content_scanner.utils.types import JsonDict

from tests.testutils import ENCRYPTED_FILE_METADATA, get_content_scanner


class EncryptedFileMetadataTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.crypto_handler = get_content_scanner().crypto_handler

    def test_unencrypted(self) -> None:
        """Tests that the _metadata_from_body function correctly returns non-encrypted
        metadata.
        """
        metadata = _metadata_from_body(ENCRYPTED_FILE_METADATA, self.crypto_handler)
        self.assertEqual(metadata, ENCRYPTED_FILE_METADATA)

    def test_encrypted(self) -> None:
        """Tests that the _metadata_from_body function correctly decrypts Olm-encrypted
        metadata and returns a decrypted version.
        """
        encrypted_body = self._encrypt_body(ENCRYPTED_FILE_METADATA)
        metadata = _metadata_from_body(encrypted_body, self.crypto_handler)
        self.assertEqual(metadata, ENCRYPTED_FILE_METADATA)

    def test_bad_json(self) -> None:
        """Tests that the _metadata_from_body function raises a REST error if the request
        body is not a valid JSON object.
        """
        with self.assertRaises(ContentScannerRestError) as cm:
            _metadata_from_body("foo", self.crypto_handler)  # type: ignore[arg-type]

        self.assertEqual(cm.exception.reason, ErrCode.MALFORMED_JSON)

    def _encrypt_body(self, content: JsonDict) -> JsonDict:
        """Encrypts the provided dictionary with Olm's PkEncryption class.

        Args:
            content: The dictionary to encrypt.

        Returns:
            An encrypted version of the dictionary in the format that's expected in POST
            requests.
        """
        msg = self.crypto_handler.encrypt(
            self.crypto_handler.public_key, json.dumps(content)
        )

        return {
            "encrypted_body": {
                "ciphertext": msg.ciphertext,
                "mac": msg.mac,
                "ephemeral": msg.ephemeral_key,
            }
        }



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import TYPE_CHECKING, Optional, Tuple

from aiohttp import web

from matrix_content_scanner.servlets import (
    _BytesResponse,
    get_media_metadata_from_request,
    web_handler,
)
from matrix_content_scanner.utils.types import JsonDict

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner


class DownloadHandler:
    def __init__(self, content_scanner: "MatrixContentScanner"):
        self._scanner = content_scanner.scanner
        self._crypto_handler = content_scanner.crypto_handler

    async def _scan(
        self,
        media_path: str,
        metadata: Optional[JsonDict] = None,
        auth_header: Optional[str] = None,
    ) -> Tuple[int, _BytesResponse]:
        media = await self._scanner.scan_file(
            media_path, metadata, auth_header=auth_header
        )

        return 200, _BytesResponse(
            headers=media.response_headers,
            content=media.content,
        )

    @web_handler
    async def handle_plain(self, request: web.Request) -> Tuple[int, _BytesResponse]:
        """Handles GET requests to ../download/serverName/mediaId"""
        media_path = request.match_info["media_path"]
        return await self._scan(
            media_path, auth_header=request.headers.get("Authorization")
        )

    @web_handler
    async def handle_encrypted(
        self, request: web.Request
    ) -> Tuple[int, _BytesResponse]:
        """Handles POST requests to ../download_encrypted"""
        media_path, metadata = await get_media_metadata_from_request(
            request, self._crypto_handler
        )

        return await self._scan(
            media_path, metadata, auth_header=request.headers.get("Authorization")
        )



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import json
import unittest

from tests.testutils import get_content_scanner


class CryptoHandlerTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.crypto_handler = get_content_scanner().crypto_handler

    def test_decrypt(self) -> None:
        """Tests that an Olm-encrypted payload is successfully decrypted."""
        payload = {"foo": "bar"}

        # Encrypt the payload with PkEncryption.
        encrypted = self.crypto_handler.encrypt(
            self.crypto_handler.public_key, json.dumps(payload)
        )

        # Decrypt the payload with the crypto handler.
        decrypted = json.loads(
            get_content_scanner().crypto_handler.decrypt_body(
                encrypted.ciphertext,
                encrypted.mac,
                encrypted.ephemeral_key,
            )
        )

        # Check that the decrypted payload is the same as the original one before
        # encryption.
        self.assertEqual(decrypted, payload)



#  Copyright 2023 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from http import HTTPStatus
from unittest.mock import patch

from aiohttp.test_utils import AioHTTPTestCase
from aiohttp.web_app import Application
from multidict import CIMultiDict

from matrix_content_scanner.httpserver import HTTPServer
from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import ContentScannerRestError

from tests.testutils import get_content_scanner

SERVER_NAME = "test"


class TestScanHandler(AioHTTPTestCase):
    def setUp(self) -> None:
        # Bypass well-known lookups.
        self.scanner = get_content_scanner(
            {"download": {"base_homeserver_url": "http://my-site.com"}}
        )

    async def get_application(self) -> Application:
        return HTTPServer(self.scanner)._app

    async def test_media_not_found_on_remote_homeserver(self) -> None:
        """Missing media on the remote HS should be presented as a 404 to the client."""
        patch_downloader = patch.object(
            self.scanner.file_downloader,
            "_get",
            return_value=(HTTPStatus.NOT_FOUND, b"", CIMultiDict()),
        )

        with patch_downloader:
            async with self.client.get(
                f"/_matrix/media_proxy/unstable/download/{SERVER_NAME}/media-does-not-exist"
            ) as resp:
                self.assertEqual(resp.status, 404)
                body = await resp.json()
                self.assertEqual(body["reason"], "M_NOT_FOUND", body)

    async def test_remote_homeserver_unreachable(self) -> None:
        """An unreachable HS should be presented as a 502 to the client."""
        patch_downloader = patch.object(
            self.scanner.file_downloader,
            "_get",
            side_effect=ContentScannerRestError(
                HTTPStatus.BAD_GATEWAY,
                ErrCode.REQUEST_FAILED,
                "dodgy network timeout :(((",
            ),
        )

        with patch_downloader:
            async with self.client.get(
                f"/_matrix/media_proxy/unstable/download/{SERVER_NAME}/media-does-not-exist"
            ) as resp:
                self.assertEqual(resp.status, 502)
                body = await resp.json()
                self.assertEqual(body["reason"], "MCS_MEDIA_REQUEST_FAILED", body)



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from enum import Enum


class ErrCode(str, Enum):
    # An unknown error happened.
    UNKNOWN = "M_UNKNOWN"
    # One of the following:
    # - No route was found with the path and method provided in the request.
    # - The homeserver does not have the requested piece of media.
    NOT_FOUND = "M_NOT_FOUND"
    # The access token is missing from the request.
    MISSING_TOKEN = "M_MISSING_TOKEN"
    # The provided access token is invalid.
    # One of the following:
    # - the access token was never valid.
    # - the access token has been logged out.
    # - the access token has been soft logged out.
    # - [Added in v1.3] the access token needs to be refreshed.
    UNKNOWN_TOKEN = "M_UNKNOWN_TOKEN"
    # The file failed the scan.
    NOT_CLEAN = "MCS_MEDIA_NOT_CLEAN"
    # The file could not be retrieved from the homeserver.
    # Does NOT cover homeserver responses with M_NOT_FOUND.
    REQUEST_FAILED = "MCS_MEDIA_REQUEST_FAILED"
    # The encrypted file could not be decrypted with the provided metadata.
    FAILED_TO_DECRYPT = "MCS_MEDIA_FAILED_TO_DECRYPT"
    # The request body isn't valid JSON, or is missing a required parameter.
    MALFORMED_JSON = "MCS_MALFORMED_JSON"
    # The Mime type is not in the allowed list of Mime types.
    MIME_TYPE_FORBIDDEN = "MCS_MIME_TYPE_FORBIDDEN"



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.

from matrix_content_scanner.utils.rust import check_rust_lib_up_to_date

check_rust_lib_up_to_date()



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import asyncio
import copy
from typing import Any, Dict, List, Optional
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, Mock

from multidict import CIMultiDict, CIMultiDictProxy

from matrix_content_scanner.scanner.scanner import CacheEntry
from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import (
    ContentScannerRestError,
    FileDirtyError,
    FileMimeTypeForbiddenError,
)
from matrix_content_scanner.utils.types import MediaDescription

from tests.testutils import (
    ENCRYPTED_FILE_METADATA,
    MEDIA_PATH,
    SMALL_BINARY_FILE,
    SMALL_PNG,
    SMALL_PNG_ENCRYPTED,
    SMALL_TEXT_FILE,
    get_content_scanner,
    to_thumbnail_params,
)


class ScannerTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.downloader_res = MediaDescription(
            content_type="image/png",
            content=SMALL_PNG,
            response_headers=CIMultiDictProxy(CIMultiDict()),
        )

        async def download_file(
            media_path: str,
            thumbnail_params: Optional[Dict[str, List[str]]] = None,
            auth_header: Optional[str] = None,
        ) -> MediaDescription:
            """Mock for the file downloader's `download_file` method."""
            return self.downloader_res

        self.downloader_mock = Mock(side_effect=download_file)

        # Mock download_file so we don't actually try to download files.
        mcs = get_content_scanner()
        mcs.file_downloader.download_file = self.downloader_mock  # type: ignore[method-assign]
        self.scanner = mcs.scanner

    async def test_scan(self) -> None:
        """Tests that we can scan files and that the scanner returns the media scanned if
        the scan was successful.
        """
        media = await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(media.content, SMALL_PNG)

    async def test_scan_dirty(self) -> None:
        """Tests that the scanner raises a FileDirtyError if the scan fails."""
        self.scanner._script = "false"
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file(MEDIA_PATH)

    async def test_encrypted_file(self) -> None:
        """Tests that the scanner can decrypt and scan encrypted files, and that if the
        scan is successful it returns the encrypted file and not the decrypted version.
        """
        self._setup_encrypted()

        media = await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)
        self.assertEqual(media.content, SMALL_PNG_ENCRYPTED)

    async def test_cache(self) -> None:
        """Tests that scan results are cached."""
        # Scan the file a first time, and check that the downloader has been called.
        await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Scan the file a second time, and check that the downloader has not been called
        # this time.
        media = await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 1)
        self.assertEqual(media.content, SMALL_PNG)

    async def test_cache_encrypted(self) -> None:
        """Tests that scan results for encrypted files are cached, and that the cached
        file is the encrypted version, not the decrypted one."""
        self._setup_encrypted()

        # Scan the file a first time, and check that the downloader has been called.
        await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Scan the file a second time, and check that the downloader has not been called
        # this time, and that the media returned is the encrypted copy.
        media = await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)
        self.assertEqual(self.downloader_mock.call_count, 1)
        self.assertEqual(media.content, SMALL_PNG_ENCRYPTED)

    async def test_cache_download_thumbnail(self) -> None:
        """Tests that cached results for full file downloads are not used for thumbnails."""
        await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 1)

        await self.scanner.scan_file(
            MEDIA_PATH, thumbnail_params=to_thumbnail_params({"width": "50"})
        )
        self.assertEqual(self.downloader_mock.call_count, 2)

    async def test_cache_thumbnail_params(self) -> None:
        """Tests that cached results for thumbnails are only used if the generation
        parameters are the same.
        """
        # Scan a thumbnail and check that the downloader was called.
        await self.scanner.scan_file(
            MEDIA_PATH, thumbnail_params=to_thumbnail_params({"width": "50"})
        )
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Scan the thumbnail again and check that the cache result was used (since the
        # downloader was not called)
        await self.scanner.scan_file(
            MEDIA_PATH, thumbnail_params=to_thumbnail_params({"width": "50"})
        )
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Scan a different thumbnail of the same media (with different parameters) and
        # check that the downloader was called.
        await self.scanner.scan_file(
            MEDIA_PATH, thumbnail_params=to_thumbnail_params({"height": "50"})
        )
        self.assertEqual(self.downloader_mock.call_count, 2)

    async def test_cache_max_size(self) -> None:
        """Tests that we don't cache files if they exceed the configured maximum file
        size.
        """
        # Set the maximum file size to be just under the size of the file.
        self.scanner._max_size_to_cache = len(SMALL_PNG) - 1

        # Scan the file a first time, and check that the downloader has been called.
        await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Scan the file a second time, and check that the downloader has been called
        # again.
        media = await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 2)
        self.assertEqual(media.content, SMALL_PNG)

    async def test_cache_max_size_mismatching_hash(self) -> None:
        """Tests that we re-scan big files if the hash we have cached for them does not
        match the hash of the newly downloaded content.
        """
        # Mock the _run_scan command so we can keep track of its call count.
        mock_runner = AsyncMock(return_value=0)
        self.scanner._run_scan = mock_runner  # type: ignore[method-assign]

        # Calculate the cache key for this file so we can look it up later.
        cache_key = self.scanner._get_cache_key_for_file(MEDIA_PATH, None, None)

        # Set the maximum file size to be just under the size of the file.
        self.scanner._max_size_to_cache = len(SMALL_PNG) - 1

        # Make sure the cache is empty.
        self.assertEqual(len(self.scanner._result_cache), 0)

        # Scan the file a first time, and check that the file has been scanned.
        await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(self.downloader_mock.call_count, 1)
        mock_runner.assert_called_once()

        # Test that the file has been cached.
        self.assertIn(cache_key, self.scanner._result_cache)

        # Change the hash of the cache entry to force it to be scanned again.
        entry: CacheEntry = self.scanner._result_cache[cache_key]
        self.scanner._result_cache[cache_key] = CacheEntry(
            result=entry.result,
            media=entry.media,
            media_hash="BAD_HASH",
            info=entry.info,
        )

        # Run the scanner again and check that the cache entry for the file has been
        # discarded (i.e. the scan is run again).
        await self.scanner.scan_file(MEDIA_PATH)
        self.assertEqual(mock_runner.call_count, 2)

        # Also check that the file has only been re-downloaded once.
        self.assertEqual(self.downloader_mock.call_count, 2)

    async def test_different_encryption_key(self) -> None:
        """Tests that if some of the file's metadata changed, we don't match against the
        cache and we download the file again.

        Also tests that the scanner fails in the correct way if it can't decrypt a file.
        """
        self._setup_encrypted()

        # Scan the file and check that the downloader was called.
        await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)
        self.assertEqual(self.downloader_mock.call_count, 1)

        # Copy the file metadata and change the key.
        modified_metadata = copy.deepcopy(ENCRYPTED_FILE_METADATA)
        modified_metadata["file"]["key"]["k"] = "somethingelse"

        # This causes the scanner to not be able to decrypt the file.
        with self.assertRaises(ContentScannerRestError) as cm:
            await self.scanner.scan_file(MEDIA_PATH, modified_metadata)

        self.assertEqual(cm.exception.http_status, 400)
        self.assertEqual(cm.exception.reason, ErrCode.FAILED_TO_DECRYPT)

        # But it also causes it to be downloaded again because its metadata have changed.
        self.assertEqual(self.downloader_mock.call_count, 2)

    async def test_allowlist_mimetype(self) -> None:
        """Tests that, if there's an allow list for MIME types and the file's MIME type
        isn't in it, the file's scan fails.
        """
        # Set an allow list that only allows JPEG files.
        self.scanner._allowed_mimetypes = ["image/jpeg"]

        # Check that the scan fails since the file is a PNG.
        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH)

    async def test_allowlist_mimetype_encrypted(self) -> None:
        """Tests that the file's MIME type is correctly detected and compared with the
        allow list (if set), even if it's encrypted.
        """
        self._setup_encrypted()

        # Set an allow list that only allows JPEG files.
        self.scanner._allowed_mimetypes = ["image/jpeg"]

        # Check that the scan fails since the file is a PNG.
        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)

    async def test_blocklist_mimetype(self) -> None:
        """Tests that, if there's an allow list for MIME types and the file's MIME type
        isn't in it, the file's scan fails.
        """
        # Set a block list that blocks PNG images.
        self.scanner._blocked_mimetypes = ["image/png"]

        # Check that the scan fails since the file is a PNG.
        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH)

    async def test_blocklist_mimetype_encrypted(self) -> None:
        """Tests that the file's MIME type is correctly detected and compared with the
        allow list (if set), even if it's encrypted.
        """
        self._setup_encrypted()

        # Set a block list that blocks PNG images.
        self.scanner._blocked_mimetypes = ["image/png"]

        # Check that the scan fails since the file is a PNG.
        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH, ENCRYPTED_FILE_METADATA)

    async def test_blocklist_mimetype_fallback_binary_file(self) -> None:
        """Tests that unrecognised binary files' MIME type is assumed to be
        `application/octet-stream` and that they can be blocked in this way.
        """

        self.downloader_res = MediaDescription(
            # This is the *claimed* content-type by the uploader
            content_type="application/vnd.io.element.generic_binary_file",
            content=SMALL_BINARY_FILE,
            response_headers=CIMultiDictProxy(CIMultiDict()),
        )

        # Set a block list that blocks uncategorised binary files.
        self.scanner._blocked_mimetypes = ["application/octet-stream"]

        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH)

    async def test_blocklist_mimetype_fallback_text_file(self) -> None:
        """Tests that unrecognised text files' MIME type is assumed to be
        `text/plain` and that they can be blocked in this way.
        """

        self.downloader_res = MediaDescription(
            # This is the *claimed* content-type by the uploader
            content_type="application/vnd.io.element.generic_file",
            content=SMALL_TEXT_FILE,
            response_headers=CIMultiDictProxy(CIMultiDict()),
        )

        # Set a block list that blocks uncategorised text files.
        self.scanner._blocked_mimetypes = ["text/plain"]

        with self.assertRaises(FileMimeTypeForbiddenError):
            await self.scanner.scan_file(MEDIA_PATH)

    async def test_dont_cache_exit_codes(self) -> None:
        """Tests that if the configuration specifies exit codes to ignore when running
        the scanning script, we don't cache them.
        """
        self.scanner._exit_codes_to_ignore = [5]

        # It's tricky to give a value to `scanner._script` that makes `_run_scan` return 5
        # directly, so we just mock it here.
        run_scan_mock = AsyncMock(return_value=5)
        self.scanner._run_scan = run_scan_mock  # type: ignore[method-assign]

        # Scan the file, we'll check later that it wasn't cached.
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file(MEDIA_PATH)

        self.assertEqual(self.downloader_mock.call_count, 1)

        # Update the mock so that the file is cached at the next scan.
        run_scan_mock.return_value = 1

        # Scan the file again to check that the file wasn't cached.
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file(MEDIA_PATH)

        self.assertEqual(self.downloader_mock.call_count, 2)

        # The file should be cached now.
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file(MEDIA_PATH)

        self.assertEqual(self.downloader_mock.call_count, 2)

    async def test_outside_temp_dir(self) -> None:
        """Tests that a scan is failed if the media path is formed in a way that would
        cause the scanner to write outside of the configured directory.
        """
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file("../bar")

    async def test_invalid_media_path(self) -> None:
        """Tests that a scan fails if the media path is invalid."""
        with self.assertRaises(FileDirtyError):
            await self.scanner.scan_file(MEDIA_PATH + "/baz")

    async def test_deduplicate_scans(self) -> None:
        """Tests that if two scan requests come in for the same file and with the same
        parameter, only one download/scan happens.
        """

        # Change the Mock's side effect to introduce some delay, to simulate a long
        # download time. We sleep asynchronously to allow additional scans requests to be
        # processed.
        async def _scan_file(*args: Any) -> MediaDescription:
            await asyncio.sleep(0.2)

            return self.downloader_res

        scan_mock = Mock(side_effect=_scan_file)
        self.scanner._scan_file = scan_mock  # type: ignore[method-assign]

        # Request two scans of the same file at the same time.
        results = await asyncio.gather(
            asyncio.create_task(self.scanner.scan_file(MEDIA_PATH)),
            asyncio.create_task(self.scanner.scan_file(MEDIA_PATH)),
        )

        # Check that the scanner has only been called once, meaning that the second
        # call did not trigger a scan.
        scan_mock.assert_called_once()

        # Check that we got two results, and that we actually got the correct media
        # description in the second scan.
        self.assertEqual(len(results), 2, results)
        self.assertEqual(results[0].content, results[1].content, results)

    def _setup_encrypted(self) -> None:
        """Sets up class properties to make the downloader return an encrypted file
        instead of a plain text one.
        """
        self.downloader_res.content_type = "application/octet-stream"
        self.downloader_res.content = SMALL_PNG_ENCRYPTED



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import json
import logging
import urllib.parse
from http import HTTPStatus
from typing import TYPE_CHECKING, Dict, Optional, Tuple

import aiohttp
from multidict import CIMultiDictProxy, MultiMapping

from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import (
    ContentScannerRestError,
    WellKnownDiscoveryError,
)
from matrix_content_scanner.utils.types import MediaDescription

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner

logger = logging.getLogger(__name__)


class _PathNotFoundException(Exception):
    """An exception raised to signal that a URL could not be found on the remote
    homeserver.
    """


class FileDownloader:
    MEDIA_DOWNLOAD_PREFIX = "_matrix/media/%s/download"
    MEDIA_THUMBNAIL_PREFIX = "_matrix/media/%s/thumbnail"
    MEDIA_DOWNLOAD_AUTHENTICATED_PREFIX = "_matrix/client/%s/media/download"
    MEDIA_THUMBNAIL_AUTHENTICATED_PREFIX = "_matrix/client/%s/media/thumbnail"

    def __init__(self, mcs: "MatrixContentScanner"):
        self._base_url = mcs.config.download.base_homeserver_url
        self._well_known_cache: Dict[str, Optional[str]] = {}
        self._proxy_url = mcs.config.download.proxy
        self._headers = (
            mcs.config.download.additional_headers
            if mcs.config.download.additional_headers is not None
            else {}
        )

    async def download_file(
        self,
        media_path: str,
        thumbnail_params: Optional[MultiMapping[str]] = None,
        auth_header: Optional[str] = None,
    ) -> MediaDescription:
        """Retrieve the file with the given `server_name/media_id` path, and stores it on
        disk.

        Args:
            media_path: The path identifying the media to retrieve.
            thumbnail_params: If present, then we want to request and scan a thumbnail
                generated with the provided parameters instead of the full media.
            auth_header: If present, we forward the given Authorization header, this is
                required for authenticated media endpoints.

        Returns:
            A description of the file (including its full content).

        Raises:
            ContentScannerRestError: The file was not found or could not be downloaded due
                to an error on the remote homeserver's side.
        """

        auth_media = True if auth_header is not None else False

        prefix = (
            self.MEDIA_DOWNLOAD_AUTHENTICATED_PREFIX
            if auth_media
            else self.MEDIA_DOWNLOAD_PREFIX
        )
        if thumbnail_params is not None:
            prefix = (
                self.MEDIA_THUMBNAIL_AUTHENTICATED_PREFIX
                if auth_media
                else self.MEDIA_THUMBNAIL_PREFIX
            )

        url = await self._build_https_url(
            media_path, prefix, "v1" if auth_media else "v3"
        )

        # Attempt to retrieve the file at the generated URL.
        try:
            file = await self._get_file_content(url, thumbnail_params, auth_header)
        except _PathNotFoundException:
            if auth_media:
                raise ContentScannerRestError(
                    http_status=HTTPStatus.NOT_FOUND,
                    reason=ErrCode.NOT_FOUND,
                    info="File not found",
                )

            # If the file could not be found, it might be because the homeserver hasn't
            # been upgraded to a version that supports Matrix v1.1 endpoints yet, so try
            # again with an r0 endpoint.
            logger.info("File not found, trying legacy r0 path")

            url = await self._build_https_url(media_path, prefix, "r0")

            try:
                file = await self._get_file_content(url, thumbnail_params, auth_header)
            except _PathNotFoundException:
                # If that still failed, raise an error.
                raise ContentScannerRestError(
                    http_status=HTTPStatus.NOT_FOUND,
                    reason=ErrCode.NOT_FOUND,
                    info="File not found",
                )

        return file

    async def _build_https_url(
        self,
        media_path: str,
        prefix: str,
        endpoint_version: str,
    ) -> str:
        """Turn a `server_name/media_id` path into an https:// one we can use to fetch
        the media.

        Note that if `base_homeserver_url` is set to an http URL, it will not be turned
        into an https one.

        Args:
            media_path: The media path to translate.
            endpoint_version: The version of the download endpoint to use. As of Matrix
                v1.11, this is "v1" for authenticated media. For unauthenticated media
                this is either "v3" or "r0".

        Returns:
            An https URL to use. If `base_homeserver_url` is set in the config, this
            will be used as the base of the URL.
        """
        server_name, media_id = media_path.split("/")

        # Figure out what base URL to use. If one is specified in the configuration file,
        # use it, otherwise try to discover one using .well-known. If that fails, use the
        # server name with an HTTPS scheme.
        if self._base_url is not None:
            base_url = self._base_url
        else:
            base_url = None

            try:
                base_url = await self._discover_via_well_known(server_name)
            except WellKnownDiscoveryError as e:
                # We don't catch ContentScannerRestErrors here because if one makes its
                # way up here then it likely means that trying to reach https://server_name
                # failed, in which case we're unlikely to be able to reach it again when
                # downloading the file, so we let the error escalate.
                logger.info("Failed to discover server via well-known: %s", e)

            if base_url is None:
                # base_url might be None if either .well-known discovery failed, or we
                # didn't find a .well-known file.
                base_url = "https://" + server_name

        # Build the full URL.
        path_prefix = prefix % endpoint_version
        url = "%s/%s/%s/%s" % (
            base_url,
            path_prefix,
            urllib.parse.quote(server_name),
            urllib.parse.quote(media_id),
        )

        return url

    async def _get_file_content(
        self,
        url: str,
        thumbnail_params: Optional[MultiMapping[str]],
        auth_header: Optional[str] = None,
    ) -> MediaDescription:
        """Retrieve the content of the file at a given URL.

        Args:
            url: The URL to query.
            thumbnail_params: Query parameters used if the request is for a thumbnail.
            auth_header: If present, we forward the given Authorization header, this is
                required for authenticated media endpoints.

        Returns:
            A description of the file (including its full content).

        Raises:
            _PathNotFoundException: the server returned an error that can mean the path
                of the request wasn't understood, e.g. because we requested a v3 URL but
                the server only supports r0, or the media couldn't be found.
                We raise a separate error class in this case because if the error is due
                to a v3 vs r0 path we want to retry the request on the r0 path.
            ContentScannerRestError: the server returned a non-200 status which cannot
                meant that the path wasn't understood.
        """
        code, body, headers = await self._get(
            url, query=thumbnail_params, auth_header=auth_header
        )

        logger.info("Remote server responded with %d", code)

        # If the response isn't a 200 OK, raise.
        if 200 < code:
            logger.info("Response body: %s", body)
            # If the response is a 404 or an "unrecognised request" à la Synapse,
            # consider that we could not find the media, and that we should retry if this
            # request was directed at a v3 endpoint.
            if code == 400:
                try:
                    err = json.loads(body)
                    if err["errcode"] == "M_UNRECOGNIZED":
                        raise _PathNotFoundException
                except (json.decoder.JSONDecodeError, KeyError):
                    pass

            if code == 401:
                try:
                    err = json.loads(body)
                    if err["errcode"] == ErrCode.MISSING_TOKEN:
                        raise ContentScannerRestError(
                            HTTPStatus.UNAUTHORIZED,
                            ErrCode.MISSING_TOKEN,
                            "Access token missing from request",
                        )
                    if err["errcode"] == ErrCode.UNKNOWN_TOKEN:
                        raise ContentScannerRestError(
                            HTTPStatus.UNAUTHORIZED,
                            ErrCode.UNKNOWN_TOKEN,
                            "Invalid access token passed",
                        )
                except (json.decoder.JSONDecodeError, KeyError):
                    pass

            if code == 404:
                raise _PathNotFoundException

            raise ContentScannerRestError(
                HTTPStatus.BAD_GATEWAY,
                ErrCode.REQUEST_FAILED,
                "The remote server responded with an error",
            )

        # Check that we have the right amount of Content-Type headers (so we don't get
        # confused later when we try comparing it with the file's MIME type).
        content_type_headers = headers.getall("content-type", None)
        if content_type_headers is None or len(content_type_headers) != 1:
            raise ContentScannerRestError(
                HTTPStatus.BAD_GATEWAY,
                ErrCode.REQUEST_FAILED,
                "The remote server responded with an invalid amount of Content-Type headers",
            )

        return MediaDescription(
            content_type=content_type_headers[0],
            content=body,
            response_headers=headers,
        )

    async def _discover_via_well_known(self, domain: str) -> Optional[str]:
        """Try to discover the base URL for the given domain via .well-known client
        discovery.

        Args:
            domain: The domain to discover the base URL for.

        Returns:
            The base URL to use, or None if no .well-known client file exist for this
            domain.

        Raises:
            WellKnownDiscoveryError if an error happened during the discovery attempt.
        """
        # Check if we already have a result cached, and if so return with it straight
        # away.
        if domain in self._well_known_cache:
            logger.info("Fetching .well-known discovery result from cache")
            return self._well_known_cache[domain]

        # Attempt to download the .well-known file.
        try:
            url = f"https://{domain}/.well-known/matrix/client"
            code, body, _ = await self._get(url)
        except ContentScannerRestError:
            raise WellKnownDiscoveryError(f"Failed to reach web server at {domain}")

        if code != 200:
            if code == 404:
                # If the response status is 404, then the homeserver hasn't set up
                # .well-known discovery, in which case we tell the caller that there's
                # no base URL to use rather than raising an error.
                # The difference is that we want to cache this result here, but we don't
                # want to do that when the discovery fails due to an incorrectly set up
                # file or an unavailable homeserver, which might be fixed later on.
                logger.info(
                    ".well-known discover has not been set up for this homeserver"
                )
                self._well_known_cache[domain] = None
                return None

            raise WellKnownDiscoveryError(
                f"Server responded with non-200 status {code}"
            )

        # Try to parse the JSON content.
        try:
            parsed_body = json.loads(body)
        except json.decoder.JSONDecodeError as e:
            raise WellKnownDiscoveryError(e)

        # Check if the parsed content has a base URL in the right place.
        try:
            base_url: str = parsed_body["m.homeserver"]["base_url"]
        except (KeyError, TypeError):
            # We might get a KeyError if we're trying to reach a key that doesn't exist,
            # and we might get a TypeError if parsed_body or parsed_body["m.homeserver"]
            # isn't a dictionary.
            raise WellKnownDiscoveryError("Response did not include a usable URL")

        # Remove the trailing slash if there is one.
        if base_url.endswith("/"):
            base_url = base_url[:-1]

        # Check if the base URL is one for a working homeserver.
        url = base_url + "/_matrix/client/versions"
        try:
            code, _, _ = await self._get(url)
        except ContentScannerRestError:
            raise WellKnownDiscoveryError(
                "Base URL does not seem to point to a working homeserver"
            )

        if code != 200:
            raise WellKnownDiscoveryError(
                "Base URL does not seem to point to a working homeserver"
            )

        # Cache and return the result.
        self._well_known_cache[domain] = base_url
        return base_url

    async def _get(
        self,
        url: str,
        query: Optional[MultiMapping[str]] = None,
        auth_header: Optional[str] = None,
    ) -> Tuple[int, bytes, CIMultiDictProxy[str]]:
        """Sends a GET request to the provided URL.

        Args:
            url: The URL to send requests to.
            query: Optional parameters to use in the request's query string.
            auth_header: If present, we forward the given Authorization header, this is
                required for authenticated media endpoints.

        Returns:
            The HTTP status code, body and headers the remote server responded with.

        Raises:
            ContentScannerRestError(502) if the request failed (if the remote server
                timed out or refused the connection, etc.).
        """
        try:
            logger.info("Sending GET request to %s", url)
            async with aiohttp.ClientSession() as session:
                if auth_header is not None:
                    request_headers = {"Authorization": auth_header, **self._headers}
                else:
                    request_headers = self._headers

                async with session.get(
                    url,
                    proxy=self._proxy_url,
                    headers=request_headers,
                    params=query,
                ) as resp:
                    return resp.status, await resp.read(), resp.headers

        except Exception as e:
            logger.error(e)
            raise ContentScannerRestError(
                HTTPStatus.BAD_GATEWAY,
                ErrCode.REQUEST_FAILED,
                "Failed to reach the remote server",
            )



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import os
from binascii import unhexlify
from typing import Dict, Optional

from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy

from matrix_content_scanner.config import MatrixContentScannerConfig
from matrix_content_scanner.mcs import MatrixContentScanner
from matrix_content_scanner.utils.types import JsonDict

# The media path to use in tests.
MEDIA_PATH = "foo/bar"

# A small, unencrypted PNG.
SMALL_PNG = unhexlify(
    b"89504e470d0a1a0a0000000d4948445200000001000000010806"
    b"0000001f15c4890000000a49444154789c63000100000500010d"
    b"0a2db40000000049454e44ae426082"
)

# A small binary file without any specific format.
SMALL_BINARY_FILE = unhexlify(b"010203")

# A small text file without any specific format.
SMALL_TEXT_FILE = b"Hello world\nThis is a tiny text file"

# A small, encrypted PNG.
SMALL_PNG_ENCRYPTED = unhexlify(
    b"9fd28dd7a1d845a04948f13af104e39402c888f7b601bce313ad"
    b"bf3e2423f67d93d5e304efc147d46df511abacbb8ae7e2e8156c"
    b"2e08de86c31fdc6aa5bd4d11537e5657102a83214d13d7ff57e6"
    b"d35940f149fbd1e661a260b0b6fe465e4e0a7c8039c08d78f679"
    b"cde511be94c685eee50571858d99d0c84918381aea3e52319509"
    b"36cac1a7b2ec46c980f3c3995eaf21fc2711b0de8ff014ff5fe7"
    b"4a7fcb3515df4f1f2ceeae72d7b58bc69d56dedf31fd430ac2ce"
    b"8aee9fcb150a1af9fdee30ac26d68d3db77c1adec5f68cad78f9"
    b"ed6ef9156ba23b76e38dfd59cb077c964248f331d43147dc7fa7"
    b"b61baf7546e5edfd78347828386b64b3a1ebdff0dcd55ea57f4b"
    b"b73b06fbedff62ef8a7fd89146fd11723e739d541d07bf399837"
    b"3ed56cb9ef475bd409e590258cdb6a0cdf4871882c334c2897c4"
    b"ea0dc76748e727a71d8c2e85253b2c80667f5d98ddbcf8fb90ba"
    b"adceb6e75a2741b740dc0d084d55cc20dd7369e7041529b62ce1"
    b"59bcde9d9a0f4978093cd52dfe77107613d2bc265519177ed623"
    b"49d70517ecf4a243fb7c20db411459766785ee6f039f68383a62"
    b"375b14cdf405401dc4aabf6812d9803218544d1ccdc9339e81cb"
    b"b36acb3414e8dfb49521b89f1b6d54a712da35e45462844a622c"
    b"aa92313335d317201e1eab5f34daba5358fde87648b24868b098"
    b"505916b8bc997b19976487718835f0d54a8794e24ca19240cad1"
    b"61e0624d8df2214edd3c33ae2b5156e2ef7191d75528f9c26a89"
    b"4a"
)

# The metadata necessary to download and decrypt SMALL_PNG_ENCRYPTED
ENCRYPTED_FILE_METADATA: JsonDict = {
    "file": {
        "v": "v2",
        "key": {
            "alg": "A256CTR",
            "ext": True,
            "k": "F3miZm2vZhucJ062AuKMUwmd-O6AK0AXP29p4MKtq3Q",
            "key_ops": ["encrypt", "decrypt"],
            "kty": "oct",
        },
        "iv": "rJqtSdi3F/EAAAAAAAAAAA",
        "hashes": {"sha256": "NYvGRRQGfyWpXSUpba+ozSbehFP6kw5ZDg0xMppyX8c"},
        "url": "mxc://" + MEDIA_PATH,
    }
}


def to_thumbnail_params(params: Dict[str, str]) -> MultiDictProxy[str]:
    """Turn the given dictionary into query parameters as they'd appear when processing a
    thumbnailing request.

    Args:
        params: The raw parameters.

    Returns:
        A multidict that can be passed onto the scanner or the file downloader.
    """
    return MultiDictProxy(MultiDict(params))


def get_base_media_headers() -> CIMultiDictProxy[str]:
    """Get the base headers necessary to react to a download request for SMALL_PNG.

    Returns:
        The headers to pass onto the file downloader.
    """
    return CIMultiDictProxy(CIMultiDict({"content-type": "image/png"}))


def get_content_scanner(config: Optional[JsonDict] = None) -> MatrixContentScanner:
    """Instantiates an instance of the content scanner.

    Args:
        config: The optional provided config.
    """
    # Create the temporary directory that we'll use so the scanner doesn't complain about
    # it not existing.
    os.makedirs(os.path.abspath("temp"), exist_ok=True)

    # We define the default configuration here rather than as a constant outside of a
    # function because otherwise a test that sets its own config would have side effects
    # on the config used for other tests.
    default_config = {
        "scan": {
            "script": "true",
            "temp_directory": "temp",
        },
        "web": {
            "host": "127.0.0.1",
            "port": 8080,
        },
        "crypto": {
            "pickle_path": "mcs_pickle.txt",
            "pickle_key": "foo",
        },
    }

    if config is None:
        config = {}

    # Update the configuration provided with some default settings.
    # Note that `update` does not update nested dictionaries (only the top level), so
    # e.g. if a configuration with a `scan` section is provided it will need to include
    # all required settings in that section.
    default_config.update(config)

    parsed_config = MatrixContentScannerConfig(default_config)

    return MatrixContentScanner(parsed_config)



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from jsonschema import ValidationError, validate

from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.errors import ContentScannerRestError
from matrix_content_scanner.utils.types import JsonDict

# This is a subset of the content of an m.room.message event that includes a file, with
# only the info that we need to locate and decrypt the file.
_encrypted_file_metadata_schema = {
    "type": "object",
    "required": ["file"],
    "properties": {
        "file": {
            "type": "object",
            "required": ["v", "iv", "url", "hashes", "key"],
            "properties": {
                "v": {"const": "v2"},
                "iv": {"type": "string"},
                "url": {"type": "string"},
                "hashes": {
                    "type": "object",
                    "required": ["sha256"],
                    "properties": {
                        "sha256": {"type": "string"},
                    },
                },
                "key": {
                    "type": "object",
                    "required": ["alg", "kty", "k", "key_ops", "ext"],
                    "properties": {
                        "alg": {"const": "A256CTR"},
                        "kty": {"const": "oct"},
                        "k": {"type": "string"},
                        "key_ops": {"type": "array", "items": {"type": "string"}},
                        "ext": {"const": True},
                    },
                },
            },
        },
    },
}


def _validate(body: JsonDict) -> None:
    """Validates the schema using jsonschema, and by checking whether the `key_ops` list
    includes at least `encrypt` and `decrypt`.

    Args:
        body: The body to validate.

    Raises:
        ValidationError if the jsonschema validation failed.
        ValueError if the `key_ops` list doesn't include at least `encrypt` and `decrypt`.
    """
    validate(body, _encrypted_file_metadata_schema)

    # We don't need to worry about triggering a KeyError/TypeError here because all of
    # these keys are marked as required in the schema, so at this point we know they're
    # here.
    key_ops = body["file"]["key"]["key_ops"]
    # We need the key_ops list to at least include "encrypt" and "decrypt", but we can't
    # check this with jsonschema, so we need to do it manually.
    if not set(key_ops).issuperset({"encrypt", "decrypt"}):
        raise ValueError('key_ops must contain at least "encrypt" and "decrypt"')


def validate_encrypted_file_metadata(body: JsonDict) -> None:
    """Validates the schema of the given dictionary, and turns any validation error
    raised into a client error.

    Args:
        body: The body to validate.

    Raises:
        ContentScannerRestError(400) if the validation failed.
    """
    # Run the validation and turns any error coming out of it into a REST error.
    try:
        _validate(body)
    except ValidationError as e:
        raise ContentScannerRestError(400, ErrCode.MALFORMED_JSON, e.message)
    except ValueError as e:
        raise ContentScannerRestError(400, ErrCode.MALFORMED_JSON, str(e))



#  Copyright 2024 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.

import os
import sys
from hashlib import blake2b

import matrix_content_scanner
from matrix_content_scanner.mcs_rust import get_rust_file_digest


def check_rust_lib_up_to_date() -> None:
    """For editable installs check if the rust library is outdated and needs to
    be rebuilt.
    """

    if not _dist_is_editable():
        return

    mcs_dir = os.path.dirname(matrix_content_scanner.__file__)
    mcs_root = os.path.abspath(os.path.join(mcs_dir, "../.."))

    # Double check we've not gone into site-packages...
    if os.path.basename(mcs_root) == "site-packages":
        return

    # ... and it looks like the root of a python project.
    if not os.path.exists("pyproject.toml"):
        return

    # Get the hash of all Rust source files
    hash = _hash_rust_files_in_directory(os.path.join(mcs_root, "rust", "src"))

    if hash != get_rust_file_digest():
        raise Exception("Rust module outdated. Please rebuild using `poetry install`")


def _hash_rust_files_in_directory(directory: str) -> str:
    """Get the hash of all files in a directory (recursively)"""

    directory = os.path.abspath(directory)

    paths = []

    dirs = [directory]
    while dirs:
        dir = dirs.pop()
        with os.scandir(dir) as d:
            for entry in d:
                if entry.is_dir():
                    dirs.append(entry.path)
                else:
                    paths.append(entry.path)

    # We sort to make sure that we get a consistent and well-defined ordering.
    paths.sort()

    hasher = blake2b()

    for path in paths:
        with open(os.path.join(directory, path), "rb") as f:
            hasher.update(f.read())

    return hasher.hexdigest()


def _dist_is_editable() -> bool:
    """Is distribution an editable install?"""
    for path_item in sys.path:
        egg_link = os.path.join(path_item, "matrix_content_scanner.pth")
        if os.path.isfile(egg_link):
            return True
    return False



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import Any, Dict

import attr
from multidict import CIMultiDictProxy


@attr.s(auto_attribs=True)
class MediaDescription:
    """A description of a media."""

    content_type: str
    content: bytes
    response_headers: CIMultiDictProxy[str]
    cacheable: bool = True


# A JSON object/dictionary.
JsonDict = Dict[str, Any]



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import copy
import unittest

from matrix_content_scanner.utils.constants import ErrCode
from matrix_content_scanner.utils.encrypted_file_metadata import (
    validate_encrypted_file_metadata,
)
from matrix_content_scanner.utils.errors import ContentScannerRestError

from tests.testutils import ENCRYPTED_FILE_METADATA


class EncryptedMetadataValidationTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.metadata = copy.deepcopy(ENCRYPTED_FILE_METADATA)

    def test_validate(self) -> None:
        """Tests that valid file metadata is considered as such."""
        validate_encrypted_file_metadata(ENCRYPTED_FILE_METADATA)

    def test_key_ops_no_decrypt(self) -> None:
        """Tests that the metadata validation fails if key_ops doesn't include `decrypt`."""
        self.metadata["file"]["key"]["key_ops"] = ["encrypt"]
        self._test_fails_validation()

    def test_key_ops_no_encrypt(self) -> None:
        """Tests that the metadata validation fails if key_ops doesn't include `encrypt`."""
        self.metadata["file"]["key"]["key_ops"] = ["decrypt"]
        self._test_fails_validation()

    def test_ops_extra_values(self) -> None:
        """tests that the metadata validation does not fail if there are extra values in
        key_ops.
        """
        self.metadata["file"]["key"]["key_ops"].append("foo")
        validate_encrypted_file_metadata(self.metadata)

    def test_no_file(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file` property."""
        self.metadata = {"foo": "bar"}
        self._test_fails_validation()

    def test_no_key(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.key` property."""
        del self.metadata["file"]["key"]
        self._test_fails_validation()

    def test_no_k(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.key.k`
        property.
        """
        del self.metadata["file"]["key"]["k"]
        self._test_fails_validation()

    def test_no_ext(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.key.ext`
        property.
        """
        del self.metadata["file"]["key"]["ext"]
        self._test_fails_validation()

    def test_bad_ext(self) -> None:
        """Tests that the metadata validation fails if the `file.key.ext` property has an
        invalid value.
        """
        self.metadata["file"]["key"]["ext"] = False
        self._test_fails_validation()

    def test_bad_alg(self) -> None:
        """Tests that the metadata validation fails if the `file.key.alg` property has an
        invalid value.
        """
        self.metadata["file"]["key"]["alg"] = "bad"
        self._test_fails_validation()

    def test_bad_kty(self) -> None:
        """Tests that the metadata validation fails if the `file.key.kty` property has an
        invalid value.
        """
        self.metadata["file"]["key"]["kty"] = "bad"
        self._test_fails_validation()

    def test_no_iv(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.iv` property."""
        del self.metadata["file"]["iv"]
        self._test_fails_validation()

    def test_no_url(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.url` property."""
        del self.metadata["file"]["url"]
        self._test_fails_validation()

    def test_no_hashes(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.hashes`
        property.
        """
        del self.metadata["file"]["hashes"]
        self._test_fails_validation()

    def test_no_sha256(self) -> None:
        """Tests that the metadata validation fails if there isn't a `file.hashes.sha256`
        property.
        """
        del self.metadata["file"]["hashes"]["sha256"]
        self._test_fails_validation()

    def _test_fails_validation(self) -> None:
        """Tests that the validation fails with a REST error complaining about malformed
        JSON.
        """
        with self.assertRaises(ContentScannerRestError) as cm:
            validate_encrypted_file_metadata(self.metadata)

        self.assertEqual(cm.exception.http_status, 400)
        self.assertEqual(cm.exception.reason, ErrCode.MALFORMED_JSON)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['matrix_content_scanner',
 'matrix_content_scanner.scanner',
 'matrix_content_scanner.servlets',
 'matrix_content_scanner.utils']

package_data = \
{'': ['*'], 'matrix_content_scanner': ['mcs_rust/*']}

install_requires = \
['aiohttp>=3.8.0',
 'attrs>=19.2.0',
 'cachetools>=5.4.0',
 'canonicaljson>=1.6.3',
 'humanfriendly>=10.0',
 'jsonschema>=4.23.0',
 'python-magic>=0.4.15,<0.5',
 'pyyaml>=5.1.1',
 'setuptools_rust>=1.3']

entry_points = \
{'console_scripts': ['matrix-content-scanner = '
                     'matrix_content_scanner.mcs:main']}

setup_kwargs = {
    'name': 'matrix_content_scanner',
    'version': '1.2.0',
    'description': 'A web service for scanning media hosted by a Matrix media repository',
    'long_description': "# Matrix Content Scanner\n\nA web service for scanning media hosted on a Matrix media repository.\n\n## Installation\n\nThis project requires libmagic to be installed on the system. On Debian/Ubuntu:\n\n```commandline\nsudo apt install libmagic1\n```\n\nThen, preferably in a virtual environment, install the Matrix Content Scanner:\n\n```commandline\npip install matrix-content-scanner\n```\n\n## Usage\n\nCopy and edit the [sample configuration file](https://github.com/matrix-org/matrix-content-scanner-python/blob/main/config.sample.yaml).\nEach key is documented in this file.\n\nThen run the content scanner (from within your virtual environment if one was created):\n\n```commandline\npython -m matrix_content_scanner.mcs -c CONFIG_FILE\n```\n\nWhere `CONFIG_FILE` is the path to your configuration file.\n\n## Docker\n\nThis project provides a Docker image to run it, published as\n`vectorim/matrix-content-scanner`.\n\nTo use it, copy the [sample configuration file](/config.sample.yaml) into a dedicated\ndirectory, edit it accordingly with your requirements, and then mount this directory as\n`/data` in the image. Do not forget to also publish the port that the content scanner's\nWeb server is configured to listen on.\n\nFor example, assuming the port for the Web server is `8080`:\n\n```shell\ndocker run -p 8080:8080 -v /path/to/your/config/directory:/data vectorim/matrix-content-scanner\n```\n\n## API\n\nSee [the API documentation](/docs/api.md) for information about how clients are expected\nto interact with the Matrix Content Scanner.\n\n## Migrating from the [legacy Matrix Content Scanner](https://github.com/matrix-org/matrix-content-scanner)\n\nBecause it uses the same APIs and Olm pickle format as the legacy Matrix Content Scanner,\nthis project can be used as a drop-in replacement. The only change (apart from the\ndeployment instructions) is the configuration format:\n\n* the `server` section is renamed `web`\n* `scan.tempDirectory` is renamed `scan.temp_directory`\n* `scan.baseUrl` is renamed `download.base_homeserver_url` (and becomes optional)\n* `scan.doNotCacheExitCodes` is renamed `result_cache.exit_codes_to_ignore`\n* `scan.directDownload` is removed. Direct download always happens when `download.base_homeserver_url`\n  is absent from the configuration file, and setting a value for it will always cause files to be\n  downloaded from the server configured.\n* `proxy` is renamed `download.proxy`\n* `middleware.encryptedBody.pickleKey` is renamed `crypto.pickle_key`\n* `middleware.encryptedBody.picklePath` is renamed `crypto.pickle_path`\n* `acceptedMimeType` is renamed `scan.allowed_mimetypes`\n* `requestHeader` is renamed `download.additional_headers` and turned into a dictionary.\n\nNote that the format of the cryptographic pickle file and key are compatible between\nthis project and the legacy Matrix Content Scanner. If no file exist at that path one will\nbe created automatically.\n\n## Development\n\nIn a virtual environment with poetry (>=1.8.3) installed, run\n```shell\npoetry install\n```\n\nTo run the unit tests, you can use:\n```shell\ntox -e py\n```\n\nTo run the linters and `mypy` type checker, use `./scripts-dev/lint.sh`.\n\n\n## Releasing\n\nThe exact steps for releasing will vary; but this is an approach taken by the\nSynapse developers (assuming a Unix-like shell):\n\n 1. Set a shell variable to the version you are releasing (this just makes\n    subsequent steps easier):\n    ```shell\n    version=X.Y.Z\n    ```\n\n 2. Update `setup.cfg` so that the `version` is correct.\n\n 3. Stage the changed files and commit.\n    ```shell\n    git add -u\n    git commit -m v$version -n\n    ```\n\n 4. Push your changes.\n    ```shell\n    git push\n    ```\n\n 5. When ready, create a signed tag for the release:\n    ```shell\n    git tag -s v$version\n    ```\n    Base the tag message on the changelog.\n\n 6. Push the tag.\n    ```shell\n    git push origin tag v$version\n    ```\n\n 7. Create a *release*, based on the tag you just pushed, on GitHub or GitLab.\n\n 8. Create a source distribution and upload it to PyPI:\n    ```shell\n    python -m build\n    twine upload dist/matrix_content_scanner-$version*\n    ```\n",
    'author': 'Element Backend Team',
    'author_email': 'team-backend-synapse@element.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10.0,<4.0.0',
}
from build_rust import *
build(setup_kwargs)

setup(**setup_kwargs)



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import Optional

from matrix_content_scanner.utils.constants import ErrCode


class ContentScannerRestError(Exception):
    """An error that is converted into an error response by the REST resource."""

    def __init__(self, http_status: int, reason: ErrCode, info: Optional[str]) -> None:
        super(Exception, self).__init__(info)
        self.http_status = http_status
        self.reason = reason
        self.info = info


class FileDirtyError(ContentScannerRestError):
    """An error indicating that the file being scanned is dirty."""

    def __init__(
        self,
        info: Optional[str] = "***VIRUS DETECTED***",
        cacheable: bool = True,
    ) -> None:
        """
        Args:
            info: The info string to serve to the client.
            cacheable: Whether raising this error should be recorded as a scan failure in
                the scanner's result cache.
        """
        super(FileDirtyError, self).__init__(
            http_status=403,
            reason=ErrCode.NOT_CLEAN,
            info=info,
        )

        self.cacheable = cacheable


class FileMimeTypeForbiddenError(ContentScannerRestError):
    """An error indicating that the file's MIME type is forbidden."""

    def __init__(self, info: Optional[str]) -> None:
        super(FileMimeTypeForbiddenError, self).__init__(
            http_status=403,
            reason=ErrCode.MIME_TYPE_FORBIDDEN,
            info=info,
        )


class ConfigError(Exception):
    """An error indicating an issue with the configuration file."""


class WellKnownDiscoveryError(Exception):
    """An error indicating a failure when attempting a .well-known discovery."""



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
from typing import TYPE_CHECKING, Tuple

from aiohttp import web

from matrix_content_scanner.servlets import _BytesResponse, web_handler

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner


class ThumbnailHandler:
    def __init__(self, content_scanner: "MatrixContentScanner"):
        self._scanner = content_scanner.scanner

    @web_handler
    async def handle_thumbnail(
        self, request: web.Request
    ) -> Tuple[int, _BytesResponse]:
        """Handles GET requests to .../thumbnail/serverName/mediaId"""
        media_path = request.match_info["media_path"]

        media = await self._scanner.scan_file(
            media_path=media_path,
            thumbnail_params=request.query,
            auth_header=request.headers.get("Authorization"),
        )

        return 200, _BytesResponse(
            headers=media.response_headers,
            content=media.content,
        )



# Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import argparse
import logging
import sys
from functools import cached_property

import yaml
from yaml.scanner import ScannerError

from matrix_content_scanner import logutils
from matrix_content_scanner.config import MatrixContentScannerConfig
from matrix_content_scanner.httpserver import HTTPServer
from matrix_content_scanner.mcs_rust import crypto, reset_logging_config
from matrix_content_scanner.scanner.file_downloader import FileDownloader
from matrix_content_scanner.scanner.scanner import Scanner
from matrix_content_scanner.utils.errors import ConfigError

logger = logging.getLogger(__name__)


class MatrixContentScanner:
    def __init__(
        self,
        config: MatrixContentScannerConfig,
    ) -> None:
        self.config = config

    @cached_property
    def file_downloader(self) -> FileDownloader:
        return FileDownloader(self)

    @cached_property
    def scanner(self) -> Scanner:
        return Scanner(self)

    @cached_property
    def crypto_handler(self) -> crypto.CryptoHandler:
        return crypto.CryptoHandler(
            self.config.crypto.pickle_key, self.config.crypto.pickle_path
        )

    def start(self) -> None:
        http_server = HTTPServer(self)
        http_server.start()


def setup_logging() -> None:
    """Basic logging setup."""
    # Set the format, this assumes every logger is created by
    # matrix_content_scanner.logging.getLogger and has custom request_type and
    # media_path fields set.
    log_format = "%(asctime)s - %(name)s - %(lineno)d - %(levelname)s - %(request_id)s - %(message)s"
    formatter = logging.Formatter(log_format)

    logutils.setup_custom_factory()

    # Create the handler and set the default logging level to INFO.
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    rootLogger = logging.getLogger("")
    rootLogger.setLevel(logging.INFO)
    rootLogger.addHandler(handler)

    reset_logging_config()


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A web service for scanning media hosted by a Matrix media repository."
    )
    parser.add_argument(
        "-c",
        type=argparse.FileType("r"),
        required=True,
        help="The YAML configuration file.",
    )

    args = parser.parse_args()

    # Load the configuration file.
    try:
        cfg = MatrixContentScannerConfig(yaml.safe_load(args.c))
    except (ConfigError, ScannerError) as e:
        # If there's an error reading the file, print it and exit without raising so we
        # don't confuse/annoy the user with an unnecessary stack trace.
        print("Failed to read configuration file: %s" % e, file=sys.stderr)
        sys.exit(1)

    # Create the content scanner.
    mcs = MatrixContentScanner(cfg)

    setup_logging()

    # Construct the crypto handler early on, so we can make sure we can load the Olm key
    # pair from the pickle file (or write it if it doesn't already exist).
    try:
        _ = mcs.crypto_handler
    except ConfigError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

    # Start the content scanner.
    mcs.start()


if __name__ == "__main__":
    main()



#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.
import logging
from typing import TYPE_CHECKING, Awaitable, Callable

from aiohttp import web

from matrix_content_scanner.servlets.download import DownloadHandler
from matrix_content_scanner.servlets.public_key import PublicKeyHandler
from matrix_content_scanner.servlets.scan import ScanHandler
from matrix_content_scanner.servlets.thumbnail import ThumbnailHandler

if TYPE_CHECKING:
    from matrix_content_scanner.mcs import MatrixContentScanner

logger = logging.getLogger(__name__)

_MEDIA_PATH_REGEXP = r"/{media_path:.+}"

_CORS_HEADERS = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
    "Access-Control-Allow-Headers": "Origin, X-Requested-With, Content-Type, Accept, Authorization",
}


@web.middleware
async def simple_cors_middleware(
    request: web.Request,
    handler: Callable[[web.Request], Awaitable[web.StreamResponse]],
) -> web.StreamResponse:
    """A simple aiohttp middleware that adds CORS headers to responses, and handles
    OPTIONS requests.

    Args:
        request: The request to handle.
        handler: The handler for this request.

    Returns:
        A response with CORS headers.
    """
    if request.method == "OPTIONS":
        # We don't register routes for OPTIONS requests, therefore the handler we're given
        # in this case just raises a 405 Method Not Allowed status using an exception.
        # Because we actually want to return a 200 OK with additional headers, we ignore
        # the handler and just return a new response.
        response = web.StreamResponse(
            status=200,
            headers=_CORS_HEADERS,
        )
        return response

    # Run the request's handler and append CORS headers to it.
    response = await handler(request)
    response.headers.update(_CORS_HEADERS)
    return response


@web.middleware
async def json_errors_middleware(
    request: web.Request,
    handler: Callable[[web.Request], Awaitable[web.StreamResponse]],
) -> web.StreamResponse:
    """A simple aiohttp middleware that converts 404/405 errors into Matrix JSON error.

    Args:
        request: The request to handle.
        handler: The handler for this request.

    Returns:
        The original response OR a JSON error response.
    """
    # Run the request's handler and append CORS headers to it.
    try:
        return await handler(request)
    except (web.HTTPNotFound, web.HTTPMethodNotAllowed) as ex:
        # Return the proper JSON response.
        return web.json_response(
            {"errcode": "M_UNRECOGNIZED", "error": "Unrecognized request"},
            status=ex.status,
        )


class HTTPServer:
    def __init__(self, mcs: "MatrixContentScanner"):
        self._mcs = mcs
        self._bind_address = mcs.config.web.host
        self._bind_port = mcs.config.web.port

        self._app = self._build_app()

    def _build_app(self) -> web.Application:
        """Build the aiohttp app and attach all the handlers to it.

        Returns:
            The built aiohttp application.
        """
        # First we build an application with all routes defined on the root path.
        app = web.Application()

        scan_handler = ScanHandler(self._mcs)
        download_handler = DownloadHandler(self._mcs)
        thumbnail_handler = ThumbnailHandler(self._mcs)
        public_key_handler = PublicKeyHandler(self._mcs)

        app.add_routes(
            [
                web.get("/scan" + _MEDIA_PATH_REGEXP, scan_handler.handle_plain),
                web.post("/scan_encrypted", scan_handler.handle_encrypted),
                web.get(
                    "/download" + _MEDIA_PATH_REGEXP, download_handler.handle_plain
                ),
                web.post("/download_encrypted", download_handler.handle_encrypted),
                web.get(
                    "/thumbnail" + _MEDIA_PATH_REGEXP,
                    thumbnail_handler.handle_thumbnail,
                ),
                web.get(
                    "/public_key",
                    public_key_handler.handle_public_key,
                ),
            ]
        )

        # Then we create a root application, and define the app we previously created as
        # a subapp on the base path for the content scanner API.
        root = web.Application(
            # Apply middlewares. This will also apply to subapps.
            middlewares=[
                # Handle trailing slashes.
                web.normalize_path_middleware(),
                # Handler CORS.
                simple_cors_middleware,
                # Convert unknown routes/methods into JSON errors.
                json_errors_middleware,
            ],
        )
        root.add_subapp("/_matrix/media_proxy/unstable", app)

        return root

    def start(self) -> None:
        """Start an aiohttp server serving the content scanner API."""
        logger.info("Starting listener on %s:%d", self._bind_address, self._bind_port)
        web.run_app(
            app=self._app,
            host=self._bind_address,
            port=self._bind_port,
            print=None,
        )
